//
// PonyLink Chip-to-Chip Interconnect
//
// Copyright (C) 2014  Clifford Wolf <clifford@clifford.at>
//
// Permission to use, copy, modify, and/or distribute this software for any
// purpose with or without fee is hereby granted, provided that the above
// copyright notice and this permission notice appear in all copies.
//
// THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
// WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
// ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
// ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
// OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
//

`timescale 1 ns / 1 ps

module ponylink_slave #(
	parameter SERDES_REG_IN = 1,
	parameter SERDES_REG_OUT = 1,
	parameter M2S_TDATA_WIDTH = 8,
	parameter M2S_TUSER_WIDTH = 0,
	parameter S2M_TDATA_WIDTH = 8,
	parameter S2M_TUSER_WIDTH = 0,
	parameter MASTER_RECV_DELAY = 4,
	parameter SLAVE_RECV_DELAY = 4,
	parameter MASTER_SEND_DELAY = 32,
	parameter SLAVE_SEND_DELAY = 32,
	parameter MASTER_PARBITS = 1,
	parameter SLAVE_PARBITS = 1,
	parameter MASTER_PKTLEN = 64,
	parameter SLAVE_PKTLEN = 64,
	// python timings.py 6 19 0.25 1.5
	parameter MASTER_TIMINGS = 80'h1d17100a040f0c090602,
	parameter SLAVE_TIMINGS  = 80'h05040302010907050301
) (
	input clk,
	output resetn_out,
	output linkerror,
	output linkready,
	output mode_recv,
	output mode_send,

	input  [7:0] gpio_i,
	output [7:0] gpio_o,

	input  [S2M_TDATA_WIDTH-1:0] in_tdata,
	input  [S2M_TUSER_WIDTH-1:0] in_tuser,
	input                        in_tvalid,
	input                        in_tlast,
	output                       in_tready,

	output [M2S_TDATA_WIDTH-1:0] out_tdata,
	output [M2S_TUSER_WIDTH-1:0] out_tuser,
	output                       out_tvalid,
	output                       out_tlast,
	input                        out_tready,

	input  [SLAVE_PARBITS-1:0] serdes_in,
	output [SLAVE_PARBITS-1:0] serdes_out,
	output [SLAVE_PARBITS-1:0] serdes_en
);
	wire [8:0] in_ser_tdata;
	wire in_ser_tvalid, in_ser_tready;

	wire [8:0] out_ser_tdata;
	wire out_ser_tvalid, out_ser_tready;

	generate if (S2M_TDATA_WIDTH <= 8 && S2M_TUSER_WIDTH <= 8) begin:pack_8bits
		ponylink_pack_8bits #(
			.TDATA_WIDTH(S2M_TDATA_WIDTH),
			.TUSER_WIDTH(S2M_TUSER_WIDTH)
		) packer (
			.clk(clk),
			.resetn(resetn_out),
			.tdata(in_tdata),
			.tuser(in_tuser),
			.tvalid(in_tvalid),
			.tlast(in_tlast),
			.tready(in_tready),
			.ser_tdata(in_ser_tdata),
			.ser_tvalid(in_ser_tvalid),
			.ser_tready(in_ser_tready)
		);
	end else begin:pack_generic
		ponylink_pack_generic #(
			.TDATA_WIDTH(S2M_TDATA_WIDTH),
			.TUSER_WIDTH(S2M_TUSER_WIDTH)
		) packer (
			.clk(clk),
			.resetn(resetn_out),
			.tdata(in_tdata),
			.tuser(in_tuser),
			.tvalid(in_tvalid),
			.tlast(in_tlast),
			.tready(in_tready),
			.ser_tdata(in_ser_tdata),
			.ser_tvalid(in_ser_tvalid),
			.ser_tready(in_ser_tready)
		);
	end endgenerate

	generate if (M2S_TDATA_WIDTH <= 8 && M2S_TUSER_WIDTH <= 8) begin:unpack_8bits
		ponylink_unpack_8bits #(
			.TDATA_WIDTH(M2S_TDATA_WIDTH),
			.TUSER_WIDTH(M2S_TUSER_WIDTH)
		) unpacker (
			.clk(clk),
			.resetn(resetn_out),
			.tdata(out_tdata),
			.tuser(out_tuser),
			.tvalid(out_tvalid),
			.tlast(out_tlast),
			.tready(out_tready),
			.ser_tdata(out_ser_tdata),
			.ser_tvalid(out_ser_tvalid),
			.ser_tready(out_ser_tready)
		);
	end else begin:unpack_generic
		ponylink_unpack_generic #(
			.TDATA_WIDTH(M2S_TDATA_WIDTH),
			.TUSER_WIDTH(M2S_TUSER_WIDTH)
		) unpacker (
			.clk(clk),
			.resetn(resetn_out),
			.tdata(out_tdata),
			.tuser(out_tuser),
			.tvalid(out_tvalid),
			.tlast(out_tlast),
			.tready(out_tready),
			.ser_tdata(out_ser_tdata),
			.ser_tvalid(out_ser_tvalid),
			.ser_tready(out_ser_tready)
		);
	end endgenerate

	wire [SLAVE_PARBITS-1:0] serdes_in_t;
	wire [SLAVE_PARBITS-1:0] serdes_out_t;
	wire [SLAVE_PARBITS-1:0] serdes_en_t;

	reg [SLAVE_PARBITS-1:0] serdes_in_r;
	reg [SLAVE_PARBITS-1:0] serdes_out_r;
	reg [SLAVE_PARBITS-1:0] serdes_en_r;

	ponylink_txrx #(
		.RECVRESET(1),
		.RECV_DELAY(SLAVE_RECV_DELAY),
		.SEND_DELAY(SLAVE_SEND_DELAY),
		.SEND_PKTLEN(SLAVE_PKTLEN),
		.PARBITS(SLAVE_PARBITS),
		.TIMINGS(SLAVE_TIMINGS)
	) txrx (
		.clk(clk),
		.resetn(1'b1),
		.resetn_out(resetn_out),
		.linkerror(linkerror),
		.linkready(linkready),
		.mode_recv(mode_recv),
		.mode_send(mode_send),

		.gpio_i(gpio_i),
		.gpio_o(gpio_o),

		.in_ser_tdata(in_ser_tdata),
		.in_ser_tvalid(in_ser_tvalid),
		.in_ser_tready(in_ser_tready),

		.out_ser_tdata(out_ser_tdata),
		.out_ser_tvalid(out_ser_tvalid),
		.out_ser_tready(out_ser_tready),

		.serdes_in(serdes_in_t),
		.serdes_out(serdes_out_t),
		.serdes_en(serdes_en_t)
	);
 
	generate if (SERDES_REG_IN) begin
		always @(posedge clk) serdes_in_r <= serdes_in;
		assign serdes_in_t = serdes_in_r;
	end else begin
		assign serdes_in_t = serdes_in;
	end endgenerate

	generate if (SERDES_REG_OUT) begin
		always @(posedge clk) serdes_out_r <= serdes_out_t;
		always @(posedge clk) serdes_en_r <= serdes_en_t;
		assign serdes_out = serdes_out_r, serdes_en = serdes_en_r;
	end else begin
		assign serdes_out = serdes_out_t, serdes_en = serdes_en_t;
	end endgenerate
endmodule

//
// PonyLink Chip-to-Chip Interconnect
//
// Copyright (C) 2014  Clifford Wolf <clifford@clifford.at>
//
// Permission to use, copy, modify, and/or distribute this software for any
// purpose with or without fee is hereby granted, provided that the above
// copyright notice and this permission notice appear in all copies.
//
// THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
// WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
// ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
// ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
// OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
//

`timescale 1 ns / 1 ps
// `define NO_PONYLINK_SCRAMBLE

module ponylink_pack_8bits #(
	parameter TDATA_WIDTH = 10,
	parameter TUSER_WIDTH = 0
)(
	input clk,
	input resetn,

	input  [TDATA_WIDTH-1:0] tdata,
	input  [TUSER_WIDTH-1:0] tuser,
	input                    tvalid,
	input                    tlast,
	output                   tready,

	output reg [8:0] ser_tdata,
	output reg       ser_tvalid,
	input            ser_tready
);
	reg [31:0] rng;
	reg [11:0] rng_counter;
	reg [4:0] rng_cursor;

	reg [7:0] queue_tuser, queue_tdata;
	reg queue_tvalid, queue_tlast, queue_send_tuser, queue_send_tuser2;
	assign tready = !queue_tvalid && resetn;

	always @(posedge clk) begin
		if (tready && tvalid) begin
			queue_tuser <= tuser;
			queue_tdata <= tdata;
			queue_tlast <= tlast;
			queue_send_tuser <= (queue_tuser !== tuser) && (TUSER_WIDTH > 0);
			queue_send_tuser2 <= 0;
			queue_tvalid <= 1;
		end
		if (!resetn) begin
`ifdef NO_PONYLINK_SCRAMBLE
			rng = 0;
`else
			rng = 123456789;
`endif
			rng_cursor <= 1;
			queue_tuser <= 0;
			queue_tlast <= 0;
			queue_send_tuser <= 0;
			queue_send_tuser2 <= 0;
			queue_tvalid <= 0;
			ser_tvalid <= 0;
		end else if (!ser_tvalid || ser_tready) begin
			if (rng_cursor) begin
				(* full_case, parallel_case *)
				case (1'b1)
					rng_cursor[0]: ser_tdata <= 9'h11c;
					rng_cursor[1]: ser_tdata <= rng[ 0 +: 8];
					rng_cursor[2]: ser_tdata <= rng[ 8 +: 8];
					rng_cursor[3]: ser_tdata <= rng[16 +: 8];
					rng_cursor[4]: ser_tdata <= rng[24 +: 8];
				endcase
				ser_tvalid <= 1;
				rng_counter <= 1;
				rng_cursor <= rng_cursor << 1;
			end else if (queue_tlast) begin
				ser_tvalid <= 1;
				ser_tdata <= 9'h17c;
				queue_tlast <= 0;
			end else if (queue_send_tuser) begin
				ser_tvalid <= 1;
				ser_tdata <= 9'h15c;
				queue_send_tuser <= 0;
				queue_send_tuser2 <= 1;
			end else if (queue_send_tuser2 || queue_tvalid) begin
				if (queue_send_tuser2) begin
					ser_tvalid <= 1;
					ser_tdata <= queue_tuser ^ rng[7:0];
					queue_send_tuser2 <= 0;
				end else if (queue_tvalid) begin
					ser_tvalid <= 1;
					ser_tdata <= queue_tdata ^ rng[7:0];
					queue_tvalid <= 0;
				end
				rng = rng ^ (rng << 13);
				rng = rng ^ (rng >> 17);
				rng = rng ^ (rng << 5);
				if (!rng_counter)
					rng_cursor <= 1;
				rng_counter <= rng_counter + 1'b1;
			end else
				ser_tvalid <= 0;
		end
	end
endmodule

module ponylink_unpack_8bits #(
	parameter TDATA_WIDTH = 8,
	parameter TUSER_WIDTH = 8
) (
	input clk,
	input resetn,

	output reg [TDATA_WIDTH-1:0] tdata,
	output reg [TUSER_WIDTH-1:0] tuser,
	output reg                   tvalid,
	output reg                   tlast,
	input                        tready,

	input  [8:0] ser_tdata,
	input        ser_tvalid,
	output reg   ser_tready
);
	reg store_tuser;

	reg [31:0] rng;
	reg [3:0] rng_cursor;
	reg rng_next;

	always @(posedge clk) begin
		rng_next <= 0;
		ser_tready <= 0;
		if (tready && tvalid) begin
			tlast <= 0;
			tvalid <= 0;
		end
		if (!resetn) begin
			rng = 123456789;
			rng_cursor <= 0;
			rng_next <= 0;
			tlast <= 0;
			tvalid <= 0;
			tuser <= 0;
		end else if (!ser_tready && ser_tvalid && (!tvalid || tready)) begin
			if (rng_cursor) begin
				(* full_case, parallel_case *)
				case (1'b1)
					rng_cursor[0]: rng[ 0 +: 8] = ser_tdata[7:0];
					rng_cursor[1]: rng[ 8 +: 8] = ser_tdata[7:0];
					rng_cursor[2]: rng[16 +: 8] = ser_tdata[7:0];
					rng_cursor[3]: rng[24 +: 8] = ser_tdata[7:0];
				endcase
				rng_cursor <= rng_cursor << 1;
				ser_tready <= 1;
			end else if (store_tuser) begin
				tuser <= ser_tdata[7:0] ^ rng[7:0];
				store_tuser <= 0;
				ser_tready <= 1;
				rng_next <= 1;
			end else if (ser_tdata == 9'h17c) begin
				tlast <= 1;
				ser_tready <= 1;
			end else if (ser_tdata == 9'h15c) begin
				store_tuser <= 1;
				ser_tready <= 1;
			end else if (ser_tdata == 9'h11c) begin
				rng_cursor <= 1;
				ser_tready <= 1;
			end else begin
				tdata <= ser_tdata[7:0] ^ rng[7:0];
				tvalid <= 1;
				ser_tready <= 1;
				rng_next <= 1;
			end
		end else if (rng_next) begin
			rng = rng ^ (rng << 13);
			rng = rng ^ (rng >> 17);
			rng = rng ^ (rng << 5);
		end
	end
endmodule

module ponylink_pack_generic #(
	parameter TDATA_WIDTH = 10,
	parameter TUSER_WIDTH = 0
) (
	input clk,
	input resetn,

	input  [TDATA_WIDTH-1:0] tdata,
	input  [TUSER_WIDTH-1:0] tuser,
	input                    tvalid,
	input                    tlast,
	output reg               tready,

	output reg [8:0] ser_tdata,
	output reg       ser_tvalid,
	input            ser_tready
);
	localparam TDATA_BYTES = (TDATA_WIDTH + 7) / 8;
	localparam TUSER_BYTES = (TUSER_WIDTH + 7) / 8;

	reg [TDATA_BYTES-1:0] cursor_td;
	reg [TUSER_BYTES-1:0] cursor_tu;
	reg [8*TDATA_BYTES-1:0] current_td;
	reg [8*TUSER_BYTES-1:0] current_tu;
	reg current_tl, mkseq_tu;

	integer i;
	reg [31:0] rng;
	reg [11:0] rng_counter;
	reg [4:0] rng_cursor;

	reg [TDATA_BYTES-1:0] nxt_cursor_td;
	reg [TUSER_BYTES-1:0] nxt_cursor_tu;
	reg nxt_current_tl, nxt_mkseq_tu;

	always @(posedge clk) begin
		ser_tdata <= 'bx;
		ser_tvalid <= 0;
		tready <= 0;
		if (!resetn) begin
			cursor_td = 0;
			cursor_tu = 0;
			current_tu = 0;
			current_tl = 0;
			mkseq_tu = 0;
			nxt_cursor_td = 0;
			nxt_cursor_tu = 0;
			nxt_current_tl = 0;
			nxt_mkseq_tu = 0;
`ifdef NO_PONYLINK_SCRAMBLE
			rng = 0;
`else
			rng = 123456789;
`endif
			rng_counter = 1;
			rng_cursor = 1;
		end else begin
			if (ser_tvalid && ser_tready) begin
				if (!current_tl && !mkseq_tu && !rng_cursor) begin
					rng = rng ^ (rng << 13);
					rng = rng ^ (rng >> 17);
					rng = rng ^ (rng << 5);
					if (!rng_counter)
						rng_cursor = 1;
					rng_counter = rng_counter+1;
				end else
					rng_cursor = rng_cursor << 1;
				cursor_td = nxt_cursor_td;
				cursor_tu = nxt_cursor_tu;
				current_tl = nxt_current_tl;
				mkseq_tu = nxt_mkseq_tu;
			end
			if (tvalid && tready) begin
				cursor_td = 1;
				cursor_tu = (current_tu === tuser || TUSER_WIDTH == 0) ? 0 : 1;
				mkseq_tu = (current_tu === tuser || TUSER_WIDTH == 0) ? 0 : 1;
				current_td = tdata;
				current_tu = tuser;
				current_tl = tlast;

				nxt_cursor_td = cursor_td;
				nxt_cursor_tu = cursor_tu;
				nxt_current_tl = current_tl;
				nxt_mkseq_tu = mkseq_tu;
			end
			if (rng_cursor) begin
				if (rng_cursor[0]) ser_tdata <= 9'h11c;
				if (rng_cursor[1]) ser_tdata <= rng[ 0 +: 8];
				if (rng_cursor[2]) ser_tdata <= rng[ 8 +: 8];
				if (rng_cursor[3]) ser_tdata <= rng[16 +: 8];
				if (rng_cursor[4]) ser_tdata <= rng[24 +: 8];
				ser_tvalid <= 1;
			end else
			if (current_tl) begin
				ser_tdata <= 9'h17c;
				ser_tvalid <= 1;
				nxt_current_tl = 0;
			end else
			if (mkseq_tu) begin
				ser_tdata <= 9'h15c;
				ser_tvalid <= 1;
				nxt_mkseq_tu = 0;
			end else
			if (cursor_tu) begin
				for (i = 0; i < TUSER_BYTES; i = i+1)
					if (cursor_tu[i]) begin ser_tdata <= current_tu[8*i +: 8] ^ rng[7:0]; ser_tvalid <= 1; end
				nxt_cursor_tu = cursor_tu << 1;
			end else
			if (cursor_td) begin
				for (i = 0; i < TDATA_BYTES; i = i+1)
					if (cursor_td[i]) begin ser_tdata <= current_td[8*i +: 8] ^ rng[7:0]; ser_tvalid <= 1; end
				nxt_cursor_td = cursor_td << 1;
			end else
				tready <= 1;
		end
	end
endmodule

module ponylink_unpack_generic #(
	parameter TDATA_WIDTH = 10,
	parameter TUSER_WIDTH = 0
) (
	input clk,
	input resetn,

	output [TDATA_WIDTH-1:0] tdata,
	output [TUSER_WIDTH-1:0] tuser,
	output                   tvalid,
	output                   tlast,
	input                    tready,

	input  [8:0] ser_tdata,
	input        ser_tvalid,
	output reg   ser_tready
);
	reg [TDATA_WIDTH-1:0] buffer_td [0:7];
	reg [TUSER_WIDTH-1:0] buffer_tu [0:7];
	reg                   buffer_tl [0:7];
	reg [2:0] buffer_in, buffer_out;

	localparam TDATA_BYTES = (TDATA_WIDTH + 7) / 8;
	localparam TUSER_BYTES = (TUSER_WIDTH + 7) / 8;

	reg [TDATA_BYTES-1:0] cursor_td;
	reg [TUSER_BYTES-1:0] cursor_tu;
	reg [8*TDATA_BYTES-1:0] next_td;
	reg [8*TUSER_BYTES-1:0] next_tu;
	reg next_tl;

	integer i;
	reg [31:0] rng;
	reg [3:0] rng_cursor;
	reg reset_q;

	assign tdata = buffer_td[buffer_out];
	assign tuser = buffer_tu[buffer_out];
	assign tlast = buffer_tl[buffer_out];
	assign tvalid = buffer_in != buffer_out && resetn;

	always @(posedge clk) begin
		reset_q <= !resetn;
		if (!resetn) begin
			buffer_in <= 0;
			buffer_out <= 0;
			ser_tready <= 0;

			cursor_td = 1;
			cursor_tu = 0;
			next_tl = 0;
			next_tu = 0;
			rng = 0;
			rng_cursor = 0;
		end else begin
			if (!cursor_td && !cursor_tu) begin
				buffer_td[buffer_in] <= next_td;
				buffer_tu[buffer_in] <= next_tu;
				buffer_tl[buffer_in] <= next_tl;
				buffer_in <= buffer_in + 1;
				ser_tready <= (buffer_out - buffer_in) > 3'd3;

				cursor_td = 1;
				next_tl = 0;
			end
			if (ser_tvalid && ser_tready) begin
				if (ser_tdata == 9'h11c) begin
					rng_cursor = 1;
				end else
				if (ser_tdata == 9'h15c && TUSER_BYTES) begin
					cursor_tu = 1;
				end else
				if (ser_tdata == 9'h17c) begin
					next_tl = 1;
				end else
				if (rng_cursor) begin
					if (rng_cursor[0]) rng[ 0 +: 8] = ser_tdata;
					if (rng_cursor[1]) rng[ 8 +: 8] = ser_tdata;
					if (rng_cursor[2]) rng[16 +: 8] = ser_tdata;
					if (rng_cursor[3]) rng[24 +: 8] = ser_tdata;
					rng_cursor = rng_cursor << 1;
				end else begin
					for (i = 0; i < TUSER_BYTES; i = i+1)
						if (cursor_tu[i]) next_tu[8*i +: 8] = ser_tdata ^ rng[7:0];
					for (i = 0; i < TDATA_BYTES; i = i+1)
						if (cursor_td[i]) next_td[8*i +: 8] = ser_tdata ^ rng[7:0];
					if (cursor_tu && TUSER_BYTES)
						cursor_tu = cursor_tu << 1;
					else
						cursor_td = cursor_td << 1;
					rng = rng ^ (rng << 13);
					rng = rng ^ (rng >> 17);
					rng = rng ^ (rng << 5);
				end
			end
			if (tvalid && tready) begin
				buffer_out <= buffer_out + 1;
				ser_tready <= 1;
			end
			if (reset_q)
				ser_tready <= 1;
		end
	end
endmodule

//
// PonyLink Chip-to-Chip Interconnect
//
// Copyright (C) 2014  Clifford Wolf <clifford@clifford.at>
//
// Permission to use, copy, modify, and/or distribute this software for any
// purpose with or without fee is hereby granted, provided that the above
// copyright notice and this permission notice appear in all copies.
//
// THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
// WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
// ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
// ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
// OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
//

`timescale 1 ns / 1 ps
// `define DEBUG

module ponylink_txrx #(
	parameter RECVRESET = 0,
	parameter PARBITS = 4,
	parameter RECV_DELAY = 0,
	parameter SEND_DELAY = 0,
	parameter SEND_PKTLEN = 64,
	parameter TIMINGS = 80'h4a3423120717120d0803
) (
	input clk,
	input resetn,
	output resetn_out,
	output linkerror,
	output linkready,
	output mode_recv,
	output mode_send,

	input      [7:0] gpio_i,
	output reg [7:0] gpio_o,

	input  [8:0] in_ser_tdata,
	input        in_ser_tvalid,
	output       in_ser_tready,

	output [8:0] out_ser_tdata,
	output       out_ser_tvalid,
	input        out_ser_tready,

	input  [PARBITS-1:0] serdes_in,
	output [PARBITS-1:0] serdes_out,
	output [PARBITS-1:0] serdes_en
);
	// Minimum length of recv pulses in samples
	localparam TIMING_R1 = TIMINGS[ 0 +: 8];  // 1 bit
	localparam TIMING_R2 = TIMINGS[ 8 +: 8];  // 2 bits
	localparam TIMING_R3 = TIMINGS[16 +: 8];  // 3 bits
	localparam TIMING_R4 = TIMINGS[24 +: 8];  // 4 bits
	localparam TIMING_R5 = TIMINGS[32 +: 8];  // 5 bits

	// Length of transmit pulses in samples
	localparam TIMING_T1 = TIMINGS[40 +: 8];  // 1 bit
	localparam TIMING_T2 = TIMINGS[48 +: 8];  // 2 bits
	localparam TIMING_T3 = TIMINGS[56 +: 8];  // 3 bits
	localparam TIMING_T4 = TIMINGS[64 +: 8];  // 4 bits
	localparam TIMING_T5 = TIMINGS[72 +: 8];  // 5 bits

	function integer my_clog2;
		input integer in_v;
		integer v;
		begin
			v = in_v;
			if (v > 0)
				v = v - 1;
			my_clog2 = 0;
			while (v) begin
				v = v >> 1;
				my_clog2 = my_clog2 + 1;
			end
		end
	endfunction

	localparam RECV_TIMEOUT = 200 * TIMING_R5 - 1;
	localparam RECV_TIMEOUT_BITS = my_clog2(RECV_TIMEOUT);
	localparam RECV_WAIT_BITS = my_clog2(RECV_DELAY);
	localparam SEND_WAIT_BITS = my_clog2(SEND_DELAY);

	localparam [8:0] MAGIC_RECV = RECVRESET ? 9'h9e : 9'h74;
	localparam [8:0] MAGIC_SEND = RECVRESET ? 9'h74 : 9'h9e;


	// --------------
	// Serdes encoder

	reg [8:0] send_word;
	reg send_valid = 0;
	wire send_ready;

	generate if (PARBITS == 1) begin:encode_1
		ponylink_txrx_encode_1 #(
			.TIMING_T1(TIMING_T1),
			.TIMING_T2(TIMING_T2),
			.TIMING_T3(TIMING_T3),
			.TIMING_T4(TIMING_T4),
			.TIMING_T5(TIMING_T5),
			.TIMING_BITS(my_clog2(TIMING_T1+TIMING_T5))
		) encoder (
			.clk(clk),
			.reset(!resetn_out),
			.send_word(send_word),
			.send_valid(send_valid),
			.send_ready(send_ready),
			.serdes_out(serdes_out),
			.serdes_en(serdes_en)
		);
	end else if (PARBITS == 2) begin:encode_2
		ponylink_txrx_encode_2 #(
			.TIMING_T1(TIMING_T1),
			.TIMING_T2(TIMING_T2),
			.TIMING_T3(TIMING_T3),
			.TIMING_T4(TIMING_T4),
			.TIMING_T5(TIMING_T5),
			.TIMING_BITS(my_clog2(TIMING_T1+TIMING_T5))
		) encoder (
			.clk(clk),
			.reset(!resetn_out),
			.send_word(send_word),
			.send_valid(send_valid),
			.send_ready(send_ready),
			.serdes_out(serdes_out),
			.serdes_en(serdes_en)
		);
	end else if (PARBITS == 4) begin:encode_4
		ponylink_txrx_encode_4 #(
			.TIMING_T1(TIMING_T1),
			.TIMING_T2(TIMING_T2),
			.TIMING_T3(TIMING_T3),
			.TIMING_T4(TIMING_T4),
			.TIMING_T5(TIMING_T5),
			.TIMING_BITS(my_clog2(TIMING_T1+TIMING_T5))
		) encoder (
			.clk(clk),
			.reset(!resetn_out),
			.send_word(send_word),
			.send_valid(send_valid),
			.send_ready(send_ready),
			.serdes_out(serdes_out),
			.serdes_en(serdes_en)
		);
	end endgenerate


	// --------------
	// Serdes decoder

	wire [8:0] recv_word;
	wire recv_word_en;
	wire recv_error;
	wire rstdetect;

	reg [3:0] reset_counter;
	assign resetn_out = !reset_counter;

	always @(posedge clk) begin
		if ((!RECVRESET && !resetn) || (RECVRESET && rstdetect))
			reset_counter <= ~0;
		else if (reset_counter)
			reset_counter <= reset_counter - 1'b1;
	end

	generate if (PARBITS == 1) begin:decode_1
		ponylink_txrx_decode_1 #(
			.RECVRESET(RECVRESET),
			.TIMING_R1(TIMING_R1),
			.TIMING_R2(TIMING_R2),
			.TIMING_R3(TIMING_R3),
			.TIMING_R4(TIMING_R4),
			.TIMING_R5(TIMING_R5),
			.TIMING_BITS(my_clog2(TIMING_R1+TIMING_R5))
		) decoder (
			.clk(clk),
			.reset(|serdes_en),
			.rstdetect(rstdetect),
			.recv_word(recv_word),
			.recv_word_en(recv_word_en),
			.recv_error(recv_error),
			.serdes_in(serdes_in)
		);
	end else if (PARBITS == 2) begin:decode_2
		ponylink_txrx_decode_2 #(
			.RECVRESET(RECVRESET),
			.TIMING_R1(TIMING_R1),
			.TIMING_R2(TIMING_R2),
			.TIMING_R3(TIMING_R3),
			.TIMING_R4(TIMING_R4),
			.TIMING_R5(TIMING_R5),
			.TIMING_BITS(my_clog2(TIMING_R1+TIMING_R5))
		) decoder (
			.clk(clk),
			.reset(|serdes_en),
			.rstdetect(rstdetect),
			.recv_word(recv_word),
			.recv_word_en(recv_word_en),
			.recv_error(recv_error),
			.serdes_in(serdes_in)
		);
	end else if (PARBITS == 4) begin:decode_4
		ponylink_txrx_decode_4 #(
			.RECVRESET(RECVRESET),
			.TIMING_R1(TIMING_R1),
			.TIMING_R2(TIMING_R2),
			.TIMING_R3(TIMING_R3),
			.TIMING_R4(TIMING_R4),
			.TIMING_R5(TIMING_R5),
			.TIMING_BITS(my_clog2(TIMING_R1+TIMING_R5))
		) decoder (
			.clk(clk),
			.reset(|serdes_en),
			.rstdetect(rstdetect),
			.recv_word(recv_word),
			.recv_word_en(recv_word_en),
			.recv_error(recv_error),
			.serdes_in(serdes_in)
		);
	end endgenerate


	// ------------------------------------------------
	// AXIS IN/OUT fifos

	reg [8:0] in_fifo_buffer [0:255];
	reg [7:0] in_fifo_iptr, in_fifo_optr;
	wire [7:0] in_fifo_iptr_nxt = in_fifo_iptr + 1'b1;
	wire [7:0] in_fifo_optr_nxt = in_fifo_optr + 1'b1;

	reg [8:0] out_fifo_buffer [0:255];
	reg [7:0] out_fifo_iptr, out_fifo_optr;
	wire [7:0] out_fifo_iptr_nxt = out_fifo_iptr + 1'b1;
	wire [7:0] out_fifo_optr_nxt = out_fifo_optr + 1'b1;

	wire [7:0] in_fifo_oblock;
	assign in_ser_tready = (in_fifo_iptr_nxt != in_fifo_oblock) && resetn_out;

	assign out_ser_tdata = out_fifo_buffer[out_fifo_optr];
	assign out_ser_tvalid = (out_fifo_iptr != out_fifo_optr) && resetn_out;

	always @(posedge clk) begin
		if (!resetn_out) begin
			in_fifo_iptr <= 0;
			out_fifo_optr <= 0;
		end else begin
			if (in_ser_tvalid && in_ser_tready) begin
				in_fifo_buffer[in_fifo_iptr] <= in_ser_tdata;
				in_fifo_iptr <= in_fifo_iptr_nxt;
			end

			if (out_ser_tvalid && out_ser_tready) begin
				out_fifo_optr <= out_fifo_optr_nxt;
			end
		end
	end


	// ------------------------------------------------
	// Half-duplex control

	reg hd_send_mode;
	reg hd_switch_to_send;
	reg hd_switch_to_recv;
	reg [RECV_TIMEOUT_BITS-1:0] hd_timeout;
	reg [RECV_WAIT_BITS-1:0] hd_recv_wait;
	reg [SEND_WAIT_BITS-1:0] hd_send_wait;
	reg hd_linkerror;
	wire hd_reset;

	always @(posedge clk) begin
		hd_linkerror <= 0;
		hd_timeout <= hd_send_mode && !hd_reset ? 0 : hd_timeout + 1;
		if (hd_recv_wait)
			hd_recv_wait <= hd_recv_wait-1;
		if (hd_send_wait)
			hd_send_wait <= hd_send_wait-1;
		if (!resetn_out) begin
			hd_timeout <= 0;
			hd_send_mode <= !RECVRESET;
			hd_recv_wait <= RECV_DELAY-1;
			hd_send_wait <= SEND_DELAY-1;
		end else
		if (hd_switch_to_send || (!RECVRESET && hd_timeout == RECV_TIMEOUT)) begin
			hd_timeout <= 0;
			hd_send_mode <= 1;
			hd_linkerror <= (!RECVRESET && hd_timeout == RECV_TIMEOUT);
			hd_send_wait <= SEND_DELAY-1;
		end else
		if (hd_switch_to_recv) begin
			hd_timeout <= 0;
			hd_send_mode <= 0;
			hd_recv_wait <= RECV_DELAY-1;
		end
	end

`ifdef DEBUG
	reg last_hd_send_mode;
	always @(posedge clk) begin
		if (hd_send_mode) begin
			if (send_valid && send_ready)
				$display("%s send: %03x", RECVRESET ? "S" : "M", send_word);
		end else begin
			if (recv_word_en)
				$display("%s recv: %03x%s", RECVRESET ? "S" : "M", recv_word, recv_error ? " *" : "");
		end
		if (last_hd_send_mode != hd_send_mode && !last_hd_send_mode)
			$display("%s ---------", RECVRESET ? "S" : "M");
		last_hd_send_mode <= hd_send_mode;
	end
`endif


	// ------------------------------------------------
	// Checksum generator

	wire [31:0] checksum;
	reg [8:0] checksum_word;
	reg checksum_enable;
	reg checksum_lock;

	ponylink_crc32 crc32 (
		.clk(clk),
		.crc_en(checksum_enable),
		.rst(checksum_word == 9'h1fc && checksum_enable),
		.data_in(checksum_word [8] ? ~checksum_word[7:0] : checksum_word[7:0]),
		.crc_out(checksum)
	);

	always @* begin
		checksum_word = 'bx;
		checksum_enable = 0;
		if (send_valid && send_ready) begin
			checksum_word = send_word;
			checksum_enable = !checksum_lock;
		end
		if (!hd_send_mode && recv_word_en) begin
			checksum_word = recv_word;
			checksum_enable = 1;
		end
	end


	// ------------------------------------------------
	// IN pkt engine

	localparam istate_preamble1 = 0;
	localparam istate_preamble2 = 1;
	localparam istate_seq_your  = 2;
	localparam istate_seq_mine  = 3;
	localparam istate_payload   = 4;
	localparam istate_gpio_ctrl = 5;
	localparam istate_gpio_data = 6;
	localparam istate_end_ctrl  = 7;
	localparam istate_checksum0 = 8;
	localparam istate_checksum1 = 9;
	localparam istate_checksum2 = 10;
	localparam istate_checksum3 = 11;

	reg [3:0] istate;
	reg [31:0] ichecksum;
	reg [7:0] igpiobuf;
	reg ilinkerror;

	reg [3:0] ilinkready = 0;
	reg linkerror = 0;
	reg linkready = 0;

	reg [7:0] next_sent_gpio;
	reg [7:0] next_sent_gpio_at;
	reg [8:0] sent_gpio;

	reg work_out_fifo_iptr_apply;
	reg [7:0] work_out_fifo_iptr;
	wire [7:0] work_out_fifo_iptr_nxt = work_out_fifo_iptr + 1;

	reg [7:0] peer_out_fifo_iptr;
	reg [7:0] peer_out_fifo_iptr_next;

	assign in_fifo_oblock = peer_out_fifo_iptr;

	always @(posedge clk) begin
		ilinkerror <= 0;
		linkready <= &ilinkready;
		hd_switch_to_send <= rstdetect;
		if (ilinkerror && !linkready)
			ilinkready <= 0;
		if (hd_linkerror || ilinkerror)
			linkerror <= 1;
		if (!resetn_out) begin
			out_fifo_iptr <= 0;
			peer_out_fifo_iptr <= 0;
			istate <= istate_preamble1;
			ilinkready <= 0;
			linkready <= 0;
			linkerror <= 0;
			sent_gpio <= 'h100;
			gpio_o <= 0;
		end else if (hd_send_mode) begin
			istate <= istate_preamble1;
		end else if (recv_word_en && !hd_recv_wait && !serdes_en) begin
			if (recv_error) begin
				istate <= istate_preamble1;
				ilinkerror <= istate != istate_preamble1 && istate != istate_preamble2;
			end else
			if (recv_word == 9'h1fc) begin
				istate <= istate_preamble2;
			end else begin
				(* full_case *)
				case (istate)
					istate_preamble1: begin
					end
					istate_preamble2: begin
						if (recv_word != 9'h1fc)
							istate <= istate_preamble1;
						if (recv_word == MAGIC_RECV) begin
							istate <= istate_seq_your;
							igpiobuf <= gpio_o;
							work_out_fifo_iptr <= out_fifo_iptr;
						end
					end
					istate_seq_your: begin
						peer_out_fifo_iptr_next <= recv_word;
						istate <= recv_word[8] ? istate_preamble1 : istate_seq_mine;
						ilinkerror <= recv_word[8] && recv_word != (RECVRESET ? 9'h1fd : 9'h1fe);
					end
					istate_seq_mine: begin
						work_out_fifo_iptr_apply <= recv_word == work_out_fifo_iptr;
						istate <= recv_word[8] ? istate_preamble1 : istate_payload;
						ilinkerror <= recv_word[8];
					end
					istate_payload: begin
						if (recv_word == 9'h19c) begin
							istate <= istate_gpio_data;
						end else if (recv_word == 9'h1bc) begin
							ichecksum <= checksum;
							istate <= istate_checksum0;
						end else if (recv_word[8] && recv_word != 9'h11c && recv_word != 9'h15c && recv_word != 9'h17c) begin
							istate <= istate_preamble1;
							ilinkerror <= 1;
						end else if (work_out_fifo_iptr_nxt == out_fifo_optr || !work_out_fifo_iptr_apply) begin
							work_out_fifo_iptr_apply <= 0;
						end else begin
							out_fifo_buffer[work_out_fifo_iptr] <= recv_word;
							work_out_fifo_iptr <= work_out_fifo_iptr_nxt;
						end
					end
					istate_gpio_data: begin
						if (recv_word == 9'h1bc) begin
							ichecksum <= checksum;
							istate <= istate_checksum0;
						end else if (recv_word[8]) begin
							istate <= istate_preamble1;
							ilinkerror <= 1;
						end else
							igpiobuf <= recv_word;
					end
					istate_checksum0: begin
						if (recv_word == ichecksum[7:0]) begin
							istate <= istate_checksum1;
						end else begin
							istate <= istate_preamble1;
							ilinkerror <= 1;
						end
					end
					istate_checksum1: begin
						if (recv_word == ichecksum[15:8]) begin
							istate <= istate_checksum2;
						end else begin
							istate <= istate_preamble1;
							ilinkerror <= 1;
						end
					end
					istate_checksum2: begin
						if (recv_word == ichecksum[23:16]) begin
							istate <= istate_checksum3;
						end else begin
							istate <= istate_preamble1;
							ilinkerror <= 1;
						end
					end
					istate_checksum3: begin
						if (recv_word == ichecksum[31:24]) begin
							gpio_o <= igpiobuf;
							peer_out_fifo_iptr <= peer_out_fifo_iptr_next;
							if (work_out_fifo_iptr_apply)
								out_fifo_iptr <= work_out_fifo_iptr;
							if (sent_gpio != next_sent_gpio)
								sent_gpio <= peer_out_fifo_iptr_next == next_sent_gpio_at &&
									peer_out_fifo_iptr_next != peer_out_fifo_iptr ? next_sent_gpio : 'h100;
							if (&linkready == 0)
								ilinkready <= ilinkready + 1;
							linkerror <= 0;
						end else
							ilinkerror <= 1;
						hd_switch_to_send <= 1;
						istate <= istate_preamble1;
					end
				endcase
			end
		end
	end


	// ------------------------------------------------
	// OUT pkt engine

	localparam ostate_preamble1 = 0;
	localparam ostate_preamble2 = 1;
	localparam ostate_preamble3 = 2;
	localparam ostate_seq_your  = 3;
	localparam ostate_seq_mine  = 4;
	localparam ostate_payload   = 5;
	localparam ostate_gpio_ctrl = 6;
	localparam ostate_gpio_data = 7;
	localparam ostate_end_ctrl  = 8;
	localparam ostate_checksum0 = 9;
	localparam ostate_checksum1 = 10;
	localparam ostate_checksum2 = 11;
	localparam ostate_checksum3 = 12;
	localparam ostate_tail      = 13;
	localparam ostate_wait      = 14;

	reg [3:0] ostate;
	reg [6:0] ostate_counter;
	reg       ostate_reset;

	assign hd_reset = ostate_reset && ostate_counter == 16 && !RECVRESET;

	always @(posedge clk) begin
		hd_switch_to_recv <= 0;
		if (!resetn_out) begin
			send_valid <= 0;
			ostate_reset <= 1;
			ostate_counter <= 0;
			ostate <= ostate_preamble1;
			in_fifo_optr <= 0;
			next_sent_gpio <= 0;
			next_sent_gpio_at <= 0;
		end else if (ostate_reset) begin
			ostate <= ostate_preamble1;
			if ((!send_valid || send_ready) && !hd_send_wait) begin
				if (ostate_counter < 16) begin
					ostate_counter <= ostate_counter + 1;
					case (ostate_counter)
						0:
							send_word <= 9'hb5;
						1:
							send_word <= 9'h1fc;
						2, 3, 4, 5:
							send_word <= RECVRESET ? 9'h1fe : 9'h1fd;
						15:
							send_word <= 9'h100;
						default:
							send_word <= ostate_counter;
					endcase
					send_valid <= 1;
				end else if (RECVRESET) begin
					ostate_reset <= 0;
					send_valid <= 0;
					hd_switch_to_recv <= 1;
				end else begin
					if (hd_timeout != RECV_TIMEOUT) begin
						if (rstdetect) begin
							send_valid <= 0;
							ostate_reset <= 0;
							hd_switch_to_recv <= 1;
						end
					end else begin
						ostate_counter <= 0;
					end
					send_valid <= 0;
				end
			end
		end else if (!hd_send_mode) begin
			ostate <= ostate_preamble1;
			send_valid <= 0;
		end else if (!hd_switch_to_recv && !hd_send_wait) begin
			(* full_case *)
			case (ostate)
				ostate_preamble1: begin
					if (!send_valid || send_ready) begin
						ostate <= ostate_preamble2;
						checksum_lock <= 0;
						send_word <= 9'hb5;
						send_valid <= 1;
					end
				end
				ostate_preamble2: begin
					if (!send_valid || send_ready) begin
						ostate <= ostate_preamble3;
						send_word <= 9'h1fc;
						send_valid <= 1;
					end
				end
				ostate_preamble3: begin
					if (!send_valid || send_ready) begin
						ostate <= ostate_seq_your;
						send_word <= MAGIC_SEND;
						send_valid <= 1;
					end
				end
				ostate_seq_your: begin
					if (!send_valid || send_ready) begin
						ostate <= ostate_seq_mine;
						send_word <= out_fifo_iptr;
						send_valid <= 1;
					end
				end
				ostate_seq_mine: begin
					if (!send_valid || send_ready) begin
						in_fifo_optr <= peer_out_fifo_iptr;
						ostate <= in_fifo_iptr != peer_out_fifo_iptr ? ostate_payload :
								gpio_i !== sent_gpio ? ostate_gpio_ctrl : ostate_end_ctrl;
						ostate_counter <= 0;
						send_word <= peer_out_fifo_iptr;
						send_valid <= 1;
					end
				end
				ostate_payload: begin
					if (send_valid && send_ready) begin
						if (in_fifo_iptr != in_fifo_optr && ostate_counter < SEND_PKTLEN)
							in_fifo_optr <= in_fifo_optr_nxt;
					end
					if (!send_valid || send_ready) begin
						if (in_fifo_iptr != in_fifo_optr && ostate_counter < SEND_PKTLEN) begin
							ostate_counter <= ostate_counter + 1;
							send_word <= in_fifo_buffer[in_fifo_optr];
							send_valid <= 1;
						end else begin
							ostate <= gpio_i !== sent_gpio ? ostate_gpio_data : ostate_checksum0;
							checksum_lock <= gpio_i === sent_gpio;
							send_word <= gpio_i !== sent_gpio ? 9'h19c : 9'h1bc;
							send_valid <= 1;
						end
					end
				end
				ostate_gpio_ctrl: begin
					if (!send_valid || send_ready) begin
						ostate <= ostate_gpio_data;
						send_word <= 9'h19c;
						send_valid <= 1;
					end
				end
				ostate_gpio_data: begin
					if (!send_valid || send_ready) begin
						ostate <= ostate_end_ctrl;
						send_word <= gpio_i;
						next_sent_gpio <= gpio_i;
						next_sent_gpio_at <= in_fifo_optr;
						send_valid <= 1;
					end
				end
				ostate_end_ctrl: begin
					if (!send_valid || send_ready) begin
						ostate <= ostate_checksum0;
						checksum_lock <= 1;
						send_word <= 9'h1bc;
						send_valid <= 1;
					end
				end
				ostate_checksum0: begin
					if (!send_valid || send_ready) begin
						ostate <= ostate_checksum1;
						checksum_lock <= 1;
						send_word <= checksum[7:0];
						send_valid <= 1;
					end
				end
				ostate_checksum1: begin
					if (!send_valid || send_ready) begin
						ostate <= ostate_checksum2;
						checksum_lock <= 1;
						send_word <= checksum[15:8];
						send_valid <= 1;
					end
				end
				ostate_checksum2: begin
					if (!send_valid || send_ready) begin
						ostate <= ostate_checksum3;
						checksum_lock <= 1;
						send_word <= checksum[23:16];
						send_valid <= 1;
					end
				end
				ostate_checksum3: begin
					if (!send_valid || send_ready) begin
						ostate <= ostate_tail;
						send_word <= checksum[31:24];
						send_valid <= 1;
					end
				end
				ostate_tail: begin
					if (!send_valid || send_ready) begin
						ostate <= ostate_wait;
						send_word <= 9'h100;
						send_valid <= 1;
					end
				end
				ostate_wait: begin
					if (!send_valid || send_ready) begin
						if (!serdes_en) begin
							hd_switch_to_recv <= 1;
							ostate <= ostate_preamble1;
						end
						send_valid <= 0;
					end
				end
			endcase
		end else begin
			if (send_ready)
				send_valid <= 0;
		end
	end

	assign mode_recv = !hd_send_mode && !hd_recv_wait && !serdes_en;
	assign mode_send = hd_send_mode && !hd_switch_to_recv && !hd_send_wait;
endmodule

// =======================================================================

module ponylink_txrx_encode_1 #(
	parameter TIMING_T1 = 1,  // 1 bit
	parameter TIMING_T2 = 2,  // 2 bits
	parameter TIMING_T3 = 3,  // 3 bits
	parameter TIMING_T4 = 4,  // 4 bits
	parameter TIMING_T5 = 5,  // 5 bits
	parameter TIMING_BITS = 8
) (
	input clk,
	input reset,

	input [8:0] send_word,
	input send_valid,
	output reg send_ready,

	output reg serdes_out,
	output reg serdes_en
);
	reg send_disp;
	wire [9:0] send_bits;
	wire next_send_disp;

	reg [14:0] buffer;
	reg [14:0] buffer_en;

	reg [14:0] next_buffer;
	reg [14:0] next_buffer_en;

	reg stage2_bits;
	reg stage2_enable;
	reg stage2_shift;

	ponylink_encode_8b10b_xtra en8b10b (
		.clk(clk),
		.datain(send_word),
		.dispin(send_disp && send_word !== 'h1fc),
		.dataout(send_bits),
		.dispout(next_send_disp)
	);

	always @(posedge clk) begin
		send_ready <= 0;
		if (reset) begin
			buffer_en <= 0;
			stage2_enable <= 0;
			send_disp <= 0;
		end else begin
			next_buffer = buffer;
			next_buffer_en = buffer_en;

			if (next_buffer_en[14:5] == 0 && send_valid) begin
				next_buffer[14:5] = send_bits;
				next_buffer_en[14:5] = {10{send_valid}};
				send_disp <= next_send_disp;
				send_ready <= send_valid;
			end

			if (stage2_shift) begin
				next_buffer = next_buffer >> 1;
				next_buffer_en = next_buffer_en >> 1;
			end

			buffer <= next_buffer;
			buffer_en <= next_buffer_en;

			stage2_bits <= next_buffer[0];
			stage2_enable <= next_buffer_en[0];
		end
	end

	reg [TIMING_BITS-1:0] max_counter = TIMING_T1;
	reg [TIMING_BITS-1:0] counter = 0;
	reg lastbit = 0;

	always @(posedge clk) begin
		stage2_shift <= 0;
		if (reset) begin
			serdes_en <= 0;
			max_counter = TIMING_T1;
			counter = 0;
			lastbit = 0;
		end else begin
			if (stage2_shift) begin
				serdes_en <= stage2_enable;
				serdes_out <= stage2_bits;
				if (!stage2_enable || lastbit != stage2_bits || max_counter == TIMING_T5) begin
					max_counter = TIMING_T1;
					counter = 0;
				end else begin
					case (max_counter)
						TIMING_T1: max_counter = TIMING_T2;
						TIMING_T2: max_counter = TIMING_T3;
						TIMING_T3: max_counter = TIMING_T4;
						TIMING_T4: max_counter = TIMING_T5;
						default: begin
							max_counter = TIMING_T1;
							counter = 0;
						end
					endcase
				end
				lastbit = stage2_bits;
			end
			counter = counter + 1'b1;
			if (max_counter == counter) begin
				stage2_shift <= 1;
			end
		end
	end
endmodule

// =======================================================================

module ponylink_txrx_encode_2 #(
	parameter TIMING_T1 = 1,  // 1 bit
	parameter TIMING_T2 = 2,  // 2 bits
	parameter TIMING_T3 = 3,  // 3 bits
	parameter TIMING_T4 = 4,  // 4 bits
	parameter TIMING_T5 = 5,  // 5 bits
	parameter TIMING_BITS = 8
) (
	input clk,
	input reset,

	input [8:0] send_word,
	input send_valid,
	output reg send_ready,

	output reg [1:0] serdes_out,
	output reg [1:0] serdes_en
);
	function integer get_max;
		input integer a, b;
		begin
			get_max = a > b ? a : b;
		end
	endfunction

	function integer get_mb2;
		input integer T1, T2, T3, T4, T5;
		begin
			get_mb2 = 2 * T1;
			get_mb2 = get_max(get_mb2, T3 - T1);
			get_mb2 = get_max(get_mb2, T4 - T2);
			get_mb2 = get_max(get_mb2, T5 - T3);
			get_mb2 = get_max(get_mb2, T1 + T2 - T1);
			get_mb2 = get_max(get_mb2, T1 + T3 - T2);
			get_mb2 = get_max(get_mb2, T1 + T4 - T3);
			get_mb2 = get_max(get_mb2, T1 + T5 - T4);
		end
	endfunction

	// maximum number of encoded samples for two consecutive data bits
	localparam TIMING_MB2 = get_mb2(TIMING_T1, TIMING_T2, TIMING_T3, TIMING_T4, TIMING_T5);

	reg send_disp;
	wire [9:0] send_bits;
	wire next_send_disp;

	reg [14:0] buffer;
	reg [14:0] buffer_en;

	reg buffer_lastbit = 0;
	reg [1:0] buffer_lastlen = 0;
	reg [1:0] buffer_thislen = 0;

	reg [14:0] next_buffer;
	reg [14:0] next_buffer_en;

	reg [TIMING_MB2-1:0] stage2_bits;
	reg [TIMING_MB2-1:0] stage2_enable;
	reg stage2_shift;

	ponylink_encode_8b10b_xtra en8b10b (
		.clk(clk),
		.datain(send_word),
		.dispin(send_disp && send_word !== 'h1fc),
		.dataout(send_bits),
		.dispout(next_send_disp)
	);

	always @(posedge clk) begin
		send_ready <= 0;
		if (reset) begin
			buffer <= 0;
			buffer_en <= 0;
			buffer_lastbit = 0;
			buffer_lastlen = 0;
			buffer_thislen = 0;
			stage2_enable <= 0;
			send_disp <= 0;
		end else begin
			next_buffer = buffer;
			next_buffer_en = buffer_en;

			if (next_buffer_en[14:5] == 0 && send_valid) begin
				next_buffer[14:5] = send_bits;
				next_buffer_en[14:5] = {10{send_valid}};
				send_disp <= next_send_disp;
				send_ready <= send_valid;
			end

			if (stage2_shift) begin
				buffer_lastbit = next_buffer[1];
				buffer_lastlen = buffer_thislen;
				next_buffer = next_buffer >> 2;
				next_buffer_en = next_buffer_en >> 2;
			end

			case (next_buffer[1:0])
				2'b00: begin
					stage2_bits <= 0;
					if (buffer_lastbit == 0) begin
						case (buffer_lastlen)
							0: stage2_enable <= {TIMING_T3-TIMING_T1{|next_buffer_en[1:0]}};
							1: stage2_enable <= {TIMING_T4-TIMING_T2{|next_buffer_en[1:0]}};
							default: stage2_enable <= {TIMING_T5-TIMING_T3{|next_buffer_en[1:0]}};
						endcase
						buffer_thislen = buffer_lastlen + 2;
					end else begin
						stage2_enable <= {TIMING_T2{1'b1}};
						buffer_thislen = 1;
					end
				end
				2'b01: begin
					if (buffer_lastbit == 1) begin
						case (buffer_lastlen)
							0: stage2_bits <= {TIMING_T2-TIMING_T1{1'b1}};
							1: stage2_bits <= {TIMING_T3-TIMING_T2{1'b1}};
							2: stage2_bits <= {TIMING_T4-TIMING_T3{1'b1}};
							3: stage2_bits <= {TIMING_T5-TIMING_T4{1'b1}};
						endcase
						case (buffer_lastlen)
							0: stage2_enable <= {TIMING_T1+TIMING_T2-TIMING_T1{|next_buffer_en[1:0]}};
							1: stage2_enable <= {TIMING_T1+TIMING_T3-TIMING_T2{|next_buffer_en[1:0]}};
							2: stage2_enable <= {TIMING_T1+TIMING_T4-TIMING_T3{|next_buffer_en[1:0]}};
							3: stage2_enable <= {TIMING_T1+TIMING_T5-TIMING_T4{|next_buffer_en[1:0]}};
						endcase
						buffer_thislen = 0;
					end else begin
						stage2_bits <= {TIMING_T1{1'b1}};
						stage2_enable <= {2*TIMING_T1{1'b1}};
						buffer_thislen = 0;
					end
				end
				2'b10: begin
					if (buffer_lastbit == 0) begin
						case (buffer_lastlen)
							0: stage2_bits <= ~{TIMING_T2-TIMING_T1{1'b1}};
							1: stage2_bits <= ~{TIMING_T3-TIMING_T2{1'b1}};
							2: stage2_bits <= ~{TIMING_T4-TIMING_T3{1'b1}};
							3: stage2_bits <= ~{TIMING_T5-TIMING_T4{1'b1}};
						endcase
						case (buffer_lastlen)
							0: stage2_enable <= {TIMING_T1+TIMING_T2-TIMING_T1{|next_buffer_en[1:0]}};
							1: stage2_enable <= {TIMING_T1+TIMING_T3-TIMING_T2{|next_buffer_en[1:0]}};
							2: stage2_enable <= {TIMING_T1+TIMING_T4-TIMING_T3{|next_buffer_en[1:0]}};
							3: stage2_enable <= {TIMING_T1+TIMING_T5-TIMING_T4{|next_buffer_en[1:0]}};
						endcase
						buffer_thislen = 0;
					end else begin
						stage2_bits <= ~{TIMING_T1{1'b1}};
						stage2_enable <= {2*TIMING_T1{1'b1}};
						buffer_thislen = 0;
					end
				end
				2'b11: begin
					stage2_bits <= ~0;
					if (buffer_lastbit == 1) begin
						case (buffer_lastlen)
							0: stage2_enable <= {TIMING_T3-TIMING_T1{|next_buffer_en[1:0]}};
							1: stage2_enable <= {TIMING_T4-TIMING_T2{|next_buffer_en[1:0]}};
							default: stage2_enable <= {TIMING_T5-TIMING_T3{|next_buffer_en[1:0]}};
						endcase
						buffer_thislen = buffer_lastlen + 2;
					end else begin
						stage2_enable <= {TIMING_T2{1'b1}};
						buffer_thislen = 1;
					end
				end
			endcase

			buffer <= next_buffer;
			buffer_en <= next_buffer_en;
		end
	end

	reg [TIMING_MB2:0] queue_bits;
	reg [TIMING_MB2:0] queue_enable = 0;

	always @(posedge clk) begin
		stage2_shift <= 0;
		if (reset) begin
			serdes_en <= 0;
			queue_enable = 0;
		end else begin
			if (stage2_shift) begin
				if (queue_enable[0]) begin
					queue_bits = {stage2_bits, queue_bits[0]};
					queue_enable = {stage2_enable, queue_enable[0]};
				end else begin
					queue_bits = stage2_bits;
					queue_enable = stage2_enable;
				end
			end

			serdes_out <= queue_bits;
			serdes_en <= queue_enable;

			queue_bits = queue_bits >> 2;
			queue_enable = queue_enable >> 2;

			stage2_shift <= !queue_enable[1];
		end
	end
endmodule

// =======================================================================

module ponylink_txrx_encode_4 #(
	parameter TIMING_T1 = 1,  // 1 bit
	parameter TIMING_T2 = 2,  // 2 bits
	parameter TIMING_T3 = 3,  // 3 bits
	parameter TIMING_T4 = 4,  // 4 bits
	parameter TIMING_T5 = 5,  // 5 bits
	parameter TIMING_BITS = 8
) (
	input clk,
	input reset,

	input [8:0] send_word,
	input send_valid,
	output reg send_ready,

	output reg [3:0] serdes_out,
	output reg [3:0] serdes_en
);
	function integer get_max;
		input integer a, b;
		begin
			get_max = a > b ? a : b;
		end
	endfunction

	function integer get_mb2;
		input integer T1, T2, T3, T4, T5;
		begin
			get_mb2 = 2 * T1;
			get_mb2 = get_max(get_mb2, T3 - T1);
			get_mb2 = get_max(get_mb2, T4 - T2);
			get_mb2 = get_max(get_mb2, T5 - T3);
			get_mb2 = get_max(get_mb2, T1 + T2 - T1);
			get_mb2 = get_max(get_mb2, T1 + T3 - T2);
			get_mb2 = get_max(get_mb2, T1 + T4 - T3);
			get_mb2 = get_max(get_mb2, T1 + T5 - T4);
		end
	endfunction

	function integer get_mb4;
		input integer T1, T2, T3, T4, T5;
		begin
			get_mb4 = 2 * get_mb2(T1, T2, T3, T4, T5);
		end
	endfunction

	// maximum number of encoded samples for four consecutive data bits
	localparam TIMING_MB4 = get_mb4(TIMING_T1, TIMING_T2, TIMING_T3, TIMING_T4, TIMING_T5);

	reg send_disp;
	wire [9:0] send_bits;
	wire next_send_disp;

	reg [20:0] buffer;
	reg [20:0] buffer_en;

	reg buffer_lastbit = 0;
	reg [1:0] buffer_lastlen = 0;
	reg [1:0] buffer_thislen = 0;

	reg [20:0] next_buffer;
	reg [20:0] next_buffer_en;

	reg [TIMING_MB4-1:0] stage2_bits;
	reg [TIMING_MB4-1:0] stage2_enable;
	reg stage2_shift;

	ponylink_encode_8b10b_xtra en8b10b (
		.clk(clk),
		.datain(send_word),
		.dispin(send_disp && send_word !== 'h1fc),
		.dataout(send_bits),
		.dispout(next_send_disp)
	);

	always @(posedge clk) begin
		send_ready <= 0;
		if (reset) begin
			buffer <= 0;
			buffer_en <= 0;
			buffer_lastbit = 0;
			buffer_lastlen = 0;
			buffer_thislen = 0;
			stage2_enable <= 0;
			send_disp <= 0;
		end else begin
			next_buffer = buffer;
			next_buffer_en = buffer_en;

			if (next_buffer_en[20:11] == 0 && send_valid) begin
				if (next_buffer_en[10:9] == 0) begin
					next_buffer[20:9] = send_bits;
					next_buffer_en[20:9] = {10{send_valid}};
				end else begin
					next_buffer[20:11] = send_bits;
					next_buffer_en[20:11] = {10{send_valid}};
				end
				send_disp <= next_send_disp;
				send_ready <= send_valid;
			end

			if (stage2_shift) begin
				buffer_lastbit = next_buffer[3];
				buffer_lastlen = buffer_thislen;
				next_buffer = next_buffer >> 4;
				next_buffer_en = next_buffer_en >> 4;
			end

			case (next_buffer[3:0])
				4'b0000, 4'b1111: begin
					stage2_bits <= next_buffer[0] ? ~0 : 0;
					if (buffer_lastbit == next_buffer[0]) begin
						stage2_enable <= {TIMING_T5-TIMING_T1{|next_buffer_en[3:0]}};
						buffer_thislen = 0;
					end else begin
						stage2_enable <= {TIMING_T4{1'b1}};
						buffer_thislen = 3;
					end
				end
				4'b0001, 4'b1110: begin
					if (buffer_lastbit == next_buffer[0]) begin
						case (buffer_lastlen)
							0: stage2_bits <= {{TIMING_T3{~next_buffer[0]}}, {TIMING_T2-TIMING_T1{next_buffer[0]}}};
							1: stage2_bits <= {{TIMING_T3{~next_buffer[0]}}, {TIMING_T3-TIMING_T2{next_buffer[0]}}};
							2: stage2_bits <= {{TIMING_T3{~next_buffer[0]}}, {TIMING_T4-TIMING_T3{next_buffer[0]}}};
							3: stage2_bits <= {{TIMING_T3{~next_buffer[0]}}, {TIMING_T5-TIMING_T4{next_buffer[0]}}};
						endcase
						case (buffer_lastlen)
							0: stage2_enable <= {TIMING_T3+TIMING_T2-TIMING_T1{|next_buffer_en[3:0]}};
							1: stage2_enable <= {TIMING_T3+TIMING_T3-TIMING_T2{|next_buffer_en[3:0]}};
							2: stage2_enable <= {TIMING_T3+TIMING_T4-TIMING_T3{|next_buffer_en[3:0]}};
							3: stage2_enable <= {TIMING_T3+TIMING_T5-TIMING_T4{|next_buffer_en[3:0]}};
						endcase
					end else begin
						stage2_bits <= {{TIMING_T3{~next_buffer[0]}}, {TIMING_T1{next_buffer[0]}}};
						stage2_enable <= {TIMING_T3+TIMING_T1{1'b1}};
					end
					buffer_thislen = 2;
				end
				4'b0010, 4'b1101: begin
					if (buffer_lastbit == next_buffer[0]) begin
						case (buffer_lastlen)
							0: stage2_bits <= {{TIMING_T2{next_buffer[0]}}, {TIMING_T1{~next_buffer[0]}}, {TIMING_T2-TIMING_T1{next_buffer[0]}}};
							1: stage2_bits <= {{TIMING_T2{next_buffer[0]}}, {TIMING_T1{~next_buffer[0]}}, {TIMING_T3-TIMING_T2{next_buffer[0]}}};
							2: stage2_bits <= {{TIMING_T2{next_buffer[0]}}, {TIMING_T1{~next_buffer[0]}}, {TIMING_T4-TIMING_T3{next_buffer[0]}}};
							3: stage2_bits <= {{TIMING_T2{next_buffer[0]}}, {TIMING_T1{~next_buffer[0]}}, {TIMING_T5-TIMING_T4{next_buffer[0]}}};
						endcase
						case (buffer_lastlen)
							0: stage2_enable <= {TIMING_T2+TIMING_T1+TIMING_T2-TIMING_T1{|next_buffer_en[3:0]}};
							1: stage2_enable <= {TIMING_T2+TIMING_T1+TIMING_T3-TIMING_T2{|next_buffer_en[3:0]}};
							2: stage2_enable <= {TIMING_T2+TIMING_T1+TIMING_T4-TIMING_T3{|next_buffer_en[3:0]}};
							3: stage2_enable <= {TIMING_T2+TIMING_T1+TIMING_T5-TIMING_T4{|next_buffer_en[3:0]}};
						endcase
					end else begin
						stage2_bits <= {{TIMING_T2{next_buffer[0]}}, {TIMING_T1{~next_buffer[0]}}, {TIMING_T1{next_buffer[0]}}};
						stage2_enable <= {TIMING_T2+2*TIMING_T1{1'b1}};
					end
					buffer_thislen = 1;
				end
				4'b0011, 4'b1100: begin
					if (buffer_lastbit == next_buffer[0]) begin
						case (buffer_lastlen)
							0: stage2_bits <= {{TIMING_T2{~next_buffer[0]}}, {TIMING_T3-TIMING_T1{next_buffer[0]}}};
							1: stage2_bits <= {{TIMING_T2{~next_buffer[0]}}, {TIMING_T4-TIMING_T2{next_buffer[0]}}};
							default: stage2_bits <= {{TIMING_T2{~next_buffer[0]}}, {TIMING_T5-TIMING_T3{next_buffer[0]}}};
						endcase
						case (buffer_lastlen)
							0: stage2_enable <= {TIMING_T2+TIMING_T3-TIMING_T1{|next_buffer_en[3:0]}};
							1: stage2_enable <= {TIMING_T2+TIMING_T4-TIMING_T2{|next_buffer_en[3:0]}};
							default: stage2_enable <= {TIMING_T2+TIMING_T5-TIMING_T3{|next_buffer_en[3:0]}};
						endcase
					end else begin
						stage2_bits <= {{TIMING_T2{~next_buffer[0]}}, {TIMING_T2{next_buffer[0]}}};
						stage2_enable <= {2*TIMING_T2{1'b1}};
					end
					buffer_thislen = 1;
				end
				4'b0100, 4'b1011: begin
					if (buffer_lastbit == next_buffer[0]) begin
						case (buffer_lastlen)
							0: stage2_bits <= {{TIMING_T1{next_buffer[0]}}, {TIMING_T1{~next_buffer[0]}}, {TIMING_T3-TIMING_T1{next_buffer[0]}}};
							1: stage2_bits <= {{TIMING_T1{next_buffer[0]}}, {TIMING_T1{~next_buffer[0]}}, {TIMING_T4-TIMING_T2{next_buffer[0]}}};
							default: stage2_bits <= {{TIMING_T1{next_buffer[0]}}, {TIMING_T1{~next_buffer[0]}}, {TIMING_T5-TIMING_T3{next_buffer[0]}}};
						endcase
						case (buffer_lastlen)
							0: stage2_enable <= {TIMING_T1+TIMING_T1+TIMING_T3-TIMING_T1{|next_buffer_en[3:0]}};
							1: stage2_enable <= {TIMING_T1+TIMING_T1+TIMING_T4-TIMING_T2{|next_buffer_en[3:0]}};
							default: stage2_enable <= {TIMING_T1+TIMING_T1+TIMING_T5-TIMING_T3{|next_buffer_en[3:0]}};
						endcase
					end else begin
						stage2_bits <= {{TIMING_T1{next_buffer[0]}}, {TIMING_T1{~next_buffer[0]}}, {TIMING_T2{next_buffer[0]}}};
						stage2_enable <= {2*TIMING_T1+TIMING_T2{1'b1}};
					end
					buffer_thislen = 0;
				end
				4'b0101, 4'b1010: begin
					if (buffer_lastbit == next_buffer[0]) begin
						case (buffer_lastlen)
							0: stage2_bits <= {{TIMING_T1{~next_buffer[0]}}, {TIMING_T1{next_buffer[0]}}, {TIMING_T1{~next_buffer[0]}}, {TIMING_T2-TIMING_T1{next_buffer[0]}}};
							1: stage2_bits <= {{TIMING_T1{~next_buffer[0]}}, {TIMING_T1{next_buffer[0]}}, {TIMING_T1{~next_buffer[0]}}, {TIMING_T3-TIMING_T2{next_buffer[0]}}};
							2: stage2_bits <= {{TIMING_T1{~next_buffer[0]}}, {TIMING_T1{next_buffer[0]}}, {TIMING_T1{~next_buffer[0]}}, {TIMING_T4-TIMING_T3{next_buffer[0]}}};
							3: stage2_bits <= {{TIMING_T1{~next_buffer[0]}}, {TIMING_T1{next_buffer[0]}}, {TIMING_T1{~next_buffer[0]}}, {TIMING_T5-TIMING_T4{next_buffer[0]}}};
						endcase
						case (buffer_lastlen)
							0: stage2_enable <= {3*TIMING_T1+TIMING_T2-TIMING_T1{|next_buffer_en[3:0]}};
							1: stage2_enable <= {3*TIMING_T1+TIMING_T3-TIMING_T2{|next_buffer_en[3:0]}};
							2: stage2_enable <= {3*TIMING_T1+TIMING_T4-TIMING_T3{|next_buffer_en[3:0]}};
							3: stage2_enable <= {3*TIMING_T1+TIMING_T5-TIMING_T4{|next_buffer_en[3:0]}};
						endcase
					end else begin
						stage2_bits <= {{TIMING_T1{~next_buffer[0]}}, {TIMING_T1{next_buffer[0]}}, {TIMING_T1{~next_buffer[0]}}, {TIMING_T1{next_buffer[0]}}};
						stage2_enable <= {4*TIMING_T1{1'b1}};
					end
					buffer_thislen = 0;
				end
				4'b0110, 4'b1001: begin
					if (buffer_lastbit == next_buffer[0]) begin
						case (buffer_lastlen)
							0: stage2_bits <= {{TIMING_T1{next_buffer[0]}}, {TIMING_T2{~next_buffer[0]}}, {TIMING_T2-TIMING_T1{next_buffer[0]}}};
							1: stage2_bits <= {{TIMING_T1{next_buffer[0]}}, {TIMING_T2{~next_buffer[0]}}, {TIMING_T3-TIMING_T2{next_buffer[0]}}};
							2: stage2_bits <= {{TIMING_T1{next_buffer[0]}}, {TIMING_T2{~next_buffer[0]}}, {TIMING_T4-TIMING_T3{next_buffer[0]}}};
							3: stage2_bits <= {{TIMING_T1{next_buffer[0]}}, {TIMING_T2{~next_buffer[0]}}, {TIMING_T5-TIMING_T4{next_buffer[0]}}};
						endcase
						case (buffer_lastlen)
							0: stage2_enable <= {TIMING_T1+TIMING_T2+TIMING_T2-TIMING_T1{|next_buffer_en[3:0]}};
							1: stage2_enable <= {TIMING_T1+TIMING_T2+TIMING_T3-TIMING_T2{|next_buffer_en[3:0]}};
							2: stage2_enable <= {TIMING_T1+TIMING_T2+TIMING_T4-TIMING_T3{|next_buffer_en[3:0]}};
							3: stage2_enable <= {TIMING_T1+TIMING_T2+TIMING_T5-TIMING_T4{|next_buffer_en[3:0]}};
						endcase
					end else begin
						stage2_bits <= {{TIMING_T1{next_buffer[0]}}, {TIMING_T2{~next_buffer[0]}}, {TIMING_T1{next_buffer[0]}}};
						stage2_enable <= {TIMING_T1+TIMING_T2+TIMING_T1{1'b1}};
					end
					buffer_thislen = 0;
				end
				4'b0111, 4'b1000: begin
					if (buffer_lastbit == next_buffer[0]) begin
						case (buffer_lastlen)
							0: stage2_bits <= {{TIMING_T1{~next_buffer[0]}}, {TIMING_T4-TIMING_T1{next_buffer[0]}}};
							default: stage2_bits <= {{TIMING_T1{~next_buffer[0]}}, {TIMING_T5-TIMING_T2{next_buffer[0]}}};
						endcase
						case (buffer_lastlen)
							0: stage2_enable <= {TIMING_T1+TIMING_T4-TIMING_T1{|next_buffer_en[3:0]}};
							default: stage2_enable <= {TIMING_T1+TIMING_T5-TIMING_T2{|next_buffer_en[3:0]}};
						endcase
					end else begin
						stage2_bits <= {{TIMING_T1{~next_buffer[0]}}, {TIMING_T3{next_buffer[0]}}};
						stage2_enable <= {TIMING_T1+TIMING_T3{1'b1}};
					end
					buffer_thislen = 0;
				end
			endcase

			buffer <= next_buffer;
			buffer_en <= next_buffer_en;
		end
	end

	reg [TIMING_MB4+3:0] queue_bits;
	reg [TIMING_MB4+3:0] queue_enable = 0;

	always @(posedge clk) begin
		stage2_shift <= 0;
		if (reset) begin
			serdes_en <= 0;
			queue_enable = 0;
		end else begin
			if (stage2_shift) begin
				if (queue_enable[2]) begin
					queue_bits = {stage2_bits, queue_bits[2:0]};
					queue_enable = {stage2_enable, queue_enable[2:0]};
				end else if (queue_enable[1]) begin
					queue_bits = {stage2_bits, queue_bits[1:0]};
					queue_enable = {stage2_enable, queue_enable[1:0]};
				end else if (queue_enable[0]) begin
					queue_bits = {stage2_bits, queue_bits[0]};
					queue_enable = {stage2_enable, queue_enable[0]};
				end else begin
					queue_bits = stage2_bits;
					queue_enable = stage2_enable;
				end
			end

			serdes_out <= queue_bits;
			serdes_en <= queue_enable;

			queue_bits = queue_bits >> 4;
			queue_enable = queue_enable >> 4;

			stage2_shift <= !queue_enable[3];
		end
	end
endmodule

// =======================================================================

module ponylink_txrx_decode_1 #(
	parameter RECVRESET = 0,
	parameter TIMING_R1 = 0,  // 1 bit
	parameter TIMING_R2 = 0,  // 2 bits
	parameter TIMING_R3 = 0,  // 3 bits
	parameter TIMING_R4 = 0,  // 4 bits
	parameter TIMING_R5 = 0,  // 5 bits
	parameter TIMING_BITS = 8
) (
	input clk,
	input reset,
	output rstdetect,

	output [8:0] recv_word,
	output reg recv_word_en,
	output recv_error,

	input [0:0] serdes_in
);
	reg [9:0] recv_wbits;
	reg [9:0] next_recv_wbits;

	ponylink_decode_8b10b_xtra #(RECVRESET) de8b10b (
		.clk(clk),
		.reset(reset),
		.enable(recv_word_en),
		.rstdetect(rstdetect),
		.datain(recv_wbits),
		.dataout(recv_word),
		.recv_error(recv_error)
	);

	reg last_bit = 0;
	reg [TIMING_BITS-1:0] cnt = 0;
	reg [3:0] cnt2 = 0;
	reg [3:0] next_cnt2;

	always @(posedge clk) begin
		recv_word_en <= 0;
		next_cnt2 = cnt2;
		next_recv_wbits = recv_wbits;
		if (serdes_in == last_bit) begin
			if (cnt == TIMING_R2-2) begin next_recv_wbits = {serdes_in, next_recv_wbits[9:1]}; next_cnt2 = next_cnt2 + 1'b1; end
			if (cnt == TIMING_R3-2) begin next_recv_wbits = {serdes_in, next_recv_wbits[9:1]}; next_cnt2 = next_cnt2 + 1'b1; end
			if (cnt == TIMING_R4-2) begin next_recv_wbits = {serdes_in, next_recv_wbits[9:1]}; next_cnt2 = next_cnt2 + 1'b1; end
			if (cnt == TIMING_R5-2) begin next_recv_wbits = {serdes_in, next_recv_wbits[9:1]}; next_cnt2 = next_cnt2 + 1'b1; end
			cnt <= cnt + 1'b1;
		end else begin
			next_recv_wbits = {serdes_in, next_recv_wbits[9:1]};
			next_cnt2 = next_cnt2 + 1'b1;
			cnt <= 0;
		end
		if (next_recv_wbits == 10'b0001111100 || next_cnt2 == 10) begin
			next_cnt2 = 0;
			recv_word_en <= !reset;
		end
		cnt2 <= next_cnt2;
		recv_wbits <= next_recv_wbits;
		last_bit <= serdes_in;
	end
endmodule

// =======================================================================

module ponylink_txrx_decode_2 #(
	parameter RECVRESET = 0,
	parameter TIMING_R1 = 0,  // 1 bit
	parameter TIMING_R2 = 0,  // 2 bits
	parameter TIMING_R3 = 0,  // 3 bits
	parameter TIMING_R4 = 0,  // 4 bits
	parameter TIMING_R5 = 0,  // 5 bits
	parameter TIMING_BITS = 8
) (
	input clk,
	input reset,
	output rstdetect,

	output [8:0] recv_word,
	output reg recv_word_en,
	output recv_error,

	input [1:0] serdes_in
);
	reg [9:0] recv_wbits;

	ponylink_decode_8b10b_xtra #(RECVRESET) de8b10b (
		.clk(clk),
		.reset(reset),
		.enable(recv_word_en),
		.rstdetect(rstdetect),
		.datain(recv_wbits),
		.dataout(recv_word),
		.recv_error(recv_error)
	);

	reg deser_last = 0;
	reg [1:0] deser_bits, deser_en;
	reg [TIMING_BITS-1:0] deser_cnt;

	wire deser_cnt_1 = deser_cnt == TIMING_R2-1 || deser_cnt == TIMING_R3-1 || deser_cnt == TIMING_R4-1 || deser_cnt == TIMING_R5-1;
	wire deser_cnt_2 = deser_cnt == TIMING_R2-2 || deser_cnt == TIMING_R3-2 || deser_cnt == TIMING_R4-2 || deser_cnt == TIMING_R5-2;

	always @(posedge clk) begin
		case (serdes_in)
			2'b00, 2'b11: begin
				if (deser_last == serdes_in[0]) begin
					deser_bits <= serdes_in;
					deser_en[0] <= deser_cnt_1 || deser_cnt_2;
					deser_en[1] <= deser_cnt_1 && deser_cnt_2;
					deser_cnt <= deser_cnt + 2;
				end else begin
					deser_bits <= serdes_in;
					deser_en[0] <= 1;
					deser_en[1] <= TIMING_R2 == 2;
					deser_cnt <= 2;
				end
			end
			2'b01, 2'b10: begin
				if (deser_last == serdes_in[0]) begin
					if (deser_cnt_1) begin
						deser_bits <= serdes_in;
						deser_en <= 2'b11;
					end else begin
						deser_bits <= serdes_in[1];
						deser_en <= 1'b1;
					end
				end else begin
					deser_bits <= serdes_in;
					deser_en <= 2'b11;
				end
				deser_cnt <= 1;
			end
		endcase
		deser_last <= serdes_in[1];
	end

	integer i;
	reg [9:0] bitbuffer;
	reg [3:0] bitcount = 0;

	always @(posedge clk) begin
		recv_word_en <= 0;
		for (i = 0; i < 2; i = i+1) begin
			if (deser_en[i]) begin
				bitcount = bitcount + 1;
				bitbuffer = {deser_bits[i], bitbuffer[9:1]};
				if (bitcount == 10 || bitbuffer == 10'b0001111100) begin
					recv_wbits <= bitbuffer;
					recv_word_en <= !reset;
					bitcount = 0;
				end
			end
		end
	end
endmodule


// =======================================================================

module ponylink_txrx_decode_4 #(
	parameter RECVRESET = 0,
	parameter TIMING_R1 = 0,  // 1 bit
	parameter TIMING_R2 = 0,  // 2 bits
	parameter TIMING_R3 = 0,  // 3 bits
	parameter TIMING_R4 = 0,  // 4 bits
	parameter TIMING_R5 = 0,  // 5 bits
	parameter TIMING_BITS = 8
) (
	input clk,
	input reset,
	output rstdetect,

	output [8:0] recv_word,
	output reg recv_word_en,
	output recv_error,

	input [3:0] serdes_in
);
	reg [9:0] recv_wbits;

	ponylink_decode_8b10b_xtra #(RECVRESET) de8b10b (
		.clk(clk),
		.reset(reset),
		.enable(recv_word_en),
		.rstdetect(rstdetect),
		.datain(recv_wbits),
		.dataout(recv_word),
		.recv_error(recv_error)
	);

	reg deser_last = 0;
	reg [3:0] deser_bits, deser_en;
	reg [TIMING_BITS-1:0] deser_cnt;

	wire deser_cnt_1 = deser_cnt == TIMING_R2-1 || deser_cnt == TIMING_R3-1 || deser_cnt == TIMING_R4-1 || deser_cnt == TIMING_R5-1;
	wire deser_cnt_2 = deser_cnt == TIMING_R2-2 || deser_cnt == TIMING_R3-2 || deser_cnt == TIMING_R4-2 || deser_cnt == TIMING_R5-2;
	wire deser_cnt_3 = deser_cnt == TIMING_R2-3 || deser_cnt == TIMING_R3-3 || deser_cnt == TIMING_R4-3 || deser_cnt == TIMING_R5-3;
	wire deser_cnt_4 = deser_cnt == TIMING_R2-4 || deser_cnt == TIMING_R3-4 || deser_cnt == TIMING_R4-4 || deser_cnt == TIMING_R5-4;

	always @(posedge clk) begin
		case (serdes_in)
			4'b0000, 4'b1111: begin
				if (deser_last == serdes_in[0]) begin
					deser_bits <= serdes_in;
					deser_en[0] <= deser_cnt_1 + deser_cnt_2 + deser_cnt_3 + deser_cnt_4 >= 3'd1;
					deser_en[1] <= deser_cnt_1 + deser_cnt_2 + deser_cnt_3 + deser_cnt_4 >= 3'd2;
					deser_en[2] <= deser_cnt_1 + deser_cnt_2 + deser_cnt_3 + deser_cnt_4 >= 3'd3;
					deser_en[3] <= deser_cnt_1 + deser_cnt_2 + deser_cnt_3 + deser_cnt_4 >= 3'd4;
					deser_cnt <= deser_cnt + 4;
				end else begin
					deser_bits <= serdes_in;
					deser_en[0] <= 1;
					deser_en[1] <= TIMING_R2 <= 4;
					deser_en[2] <= TIMING_R3 <= 4;
					deser_en[3] <= TIMING_R4 <= 4;
					deser_cnt <= 4;
				end
			end
			4'b0001, 4'b1110: begin
				if (deser_last == serdes_in[0]) begin
					if (deser_cnt_1) begin
						deser_bits <= serdes_in;
						deser_en[0] <= 1;
						deser_en[1] <= 1;
						deser_en[2] <= TIMING_R2 <= 3;
						deser_en[3] <= TIMING_R3 <= 3;
					end else begin
						deser_bits <= serdes_in >> 1;
						deser_en[0] <= 1;
						deser_en[1] <= TIMING_R2 <= 3;
						deser_en[2] <= TIMING_R3 <= 3;
						deser_en[3] <= 0;
					end
				end else begin
					deser_bits <= serdes_in;
					deser_en[0] <= 1;
					deser_en[1] <= 1;
					deser_en[2] <= TIMING_R2 <= 3;
					deser_en[3] <= TIMING_R3 <= 3;
				end
				deser_cnt <= 3;
			end
			4'b0011, 4'b1100: begin
				if (deser_last == serdes_in[0]) begin
					if (deser_cnt_1 && deser_cnt_2) begin
						deser_bits <= serdes_in;
						deser_en[0] <= 1;
						deser_en[1] <= 1;
						deser_en[2] <= 1;
						deser_en[3] <= TIMING_R2 <= 2;
					end else
					if (deser_cnt_1 || deser_cnt_2) begin
						deser_bits <= serdes_in >> 1;
						deser_en[0] <= 1;
						deser_en[1] <= 1;
						deser_en[2] <= TIMING_R2 <= 2;
						deser_en[3] <= 0;
					end else begin
						deser_bits <= serdes_in >> 2;
						deser_en[0] <= 1;
						deser_en[1] <= TIMING_R2 <= 2;
						deser_en[2] <= 0;
						deser_en[3] <= 0;
					end
				end else begin
					if (TIMING_R2 <= 2) begin
						deser_bits <= serdes_in;
						deser_en[0] <= 1;
						deser_en[1] <= 1;
						deser_en[2] <= 1;
						deser_en[3] <= 1;
					end else begin
						deser_bits <= serdes_in >> 1;
						deser_en[0] <= 1;
						deser_en[1] <= 1;
						deser_en[2] <= 0;
						deser_en[3] <= 0;
					end
				end
				deser_cnt <= 2;
			end
			4'b0111, 4'b1000: begin
				if (deser_last == serdes_in[0]) begin
					if (deser_cnt_1 + deser_cnt_2 + deser_cnt_3 >= 3'd3) begin
						deser_bits <= serdes_in;
						deser_en[0] <= 1;
						deser_en[1] <= 1;
						deser_en[2] <= 1;
						deser_en[3] <= 1;
					end else
					if (deser_cnt_1 + deser_cnt_2 + deser_cnt_3 >= 3'd2) begin
						deser_bits <= serdes_in >> 1;
						deser_en[0] <= 1;
						deser_en[1] <= 1;
						deser_en[2] <= 1;
						deser_en[3] <= 0;
					end else
					if (deser_cnt_1 + deser_cnt_2 + deser_cnt_3 >= 3'd1) begin
						deser_bits <= serdes_in >> 2;
						deser_en[0] <= 1;
						deser_en[1] <= 1;
						deser_en[2] <= 0;
						deser_en[3] <= 0;
					end else begin
						deser_bits <= serdes_in >> 3;
						deser_en[0] <= 1;
						deser_en[1] <= 0;
						deser_en[2] <= 0;
						deser_en[3] <= 0;
					end
				end else begin
					if (TIMING_R3 <= 3) begin
						deser_bits <= serdes_in;
						deser_en[0] <= 1;
						deser_en[1] <= 1;
						deser_en[2] <= 1;
						deser_en[3] <= 1;
					end else
					if (TIMING_R2 <= 3) begin
						deser_bits <= serdes_in >> 1;
						deser_en[0] <= 1;
						deser_en[1] <= 1;
						deser_en[2] <= 1;
						deser_en[3] <= 0;
					end else begin
						deser_bits <= serdes_in >> 2;
						deser_en[0] <= 1;
						deser_en[1] <= 1;
						deser_en[2] <= 0;
						deser_en[3] <= 0;
					end
				end
				deser_cnt <= 1;
			end
			4'b0110, 4'b1001: begin
				if (deser_last == serdes_in[0]) begin
					if (deser_cnt_1) begin
						if (TIMING_R2 <= 2) begin
							deser_bits <= serdes_in;
							deser_en[0] <= 1;
							deser_en[1] <= 1;
							deser_en[2] <= 1;
							deser_en[3] <= 1;
						end else begin
							deser_bits <= {serdes_in[3], serdes_in[2], serdes_in[0]};
							deser_en[0] <= 1;
							deser_en[1] <= 1;
							deser_en[2] <= 1;
							deser_en[3] <= 0;
						end
					end else begin
						if (TIMING_R2 <= 2) begin
							deser_bits <= serdes_in >> 1;
							deser_en[0] <= 1;
							deser_en[1] <= 1;
							deser_en[2] <= 1;
							deser_en[3] <= 0;
						end else begin
							deser_bits <= serdes_in >> 2;
							deser_en[0] <= 1;
							deser_en[1] <= 1;
							deser_en[2] <= 0;
							deser_en[3] <= 0;
						end
					end
				end else begin
					if (TIMING_R2 <= 2) begin
						deser_bits <= serdes_in;
						deser_en[0] <= 1;
						deser_en[1] <= 1;
						deser_en[2] <= 1;
						deser_en[3] <= 1;
					end else begin
						deser_bits <= {serdes_in[3], serdes_in[2], serdes_in[0]};
						deser_en[0] <= 1;
						deser_en[1] <= 1;
						deser_en[2] <= 1;
						deser_en[3] <= 0;
					end
				end
				deser_cnt <= 1;
			end
			4'b0010, 4'b1101: begin
				if (deser_last == serdes_in[0]) begin
					if (deser_cnt_1) begin
						deser_bits <= serdes_in;
						deser_en[0] <= 1;
						deser_en[1] <= 1;
						deser_en[2] <= 1;
						deser_en[3] <= TIMING_R2 <= 2;
					end else begin
						deser_bits <= serdes_in >> 1;
						deser_en[0] <= 1;
						deser_en[1] <= 1;
						deser_en[2] <= TIMING_R2 <= 2;
						deser_en[3] <= 0;
					end
				end else begin
					deser_bits <= serdes_in;
					deser_en[0] <= 1;
					deser_en[1] <= 1;
					deser_en[2] <= 1;
					deser_en[3] <= TIMING_R2 <= 2;
				end
				deser_cnt <= 2;
			end
			4'b0100, 4'b1011: begin
				if (deser_last == serdes_in[0]) begin
					if (deser_cnt_1 && deser_cnt_2) begin
						deser_bits <= serdes_in;
						deser_en[0] <= 1;
						deser_en[1] <= 1;
						deser_en[2] <= 1;
						deser_en[3] <= 1;
					end else
					if (deser_cnt_1 || deser_cnt_2) begin
						deser_bits <= serdes_in >> 1;
						deser_en[0] <= 1;
						deser_en[1] <= 1;
						deser_en[2] <= 1;
						deser_en[3] <= 0;
					end else begin
						deser_bits <= serdes_in >> 2;
						deser_en[0] <= 1;
						deser_en[1] <= 1;
						deser_en[2] <= 0;
						deser_en[3] <= 0;
					end
				end else begin
					if (TIMING_R2 <= 2) begin
						deser_bits <= serdes_in;
						deser_en[0] <= 1;
						deser_en[1] <= 1;
						deser_en[2] <= 1;
						deser_en[3] <= 1;
					end else begin
						deser_bits <= serdes_in >> 1;
						deser_en[0] <= 1;
						deser_en[1] <= 1;
						deser_en[2] <= 1;
						deser_en[3] <= 0;
					end
				end
				deser_cnt <= 1;
			end
			4'b0101, 4'b1010: begin
				if (deser_last == serdes_in[0]) begin
					if (deser_cnt_1) begin
						deser_bits <= serdes_in;
						deser_en[0] <= 1;
						deser_en[1] <= 1;
						deser_en[2] <= 1;
						deser_en[3] <= 1;
					end else begin
						deser_bits <= serdes_in >> 1;
						deser_en[0] <= 1;
						deser_en[1] <= 1;
						deser_en[2] <= 1;
						deser_en[3] <= 0;
					end
				end else begin
					deser_bits <= serdes_in;
					deser_en[0] <= 1;
					deser_en[1] <= 1;
					deser_en[2] <= 1;
					deser_en[3] <= 1;
				end
				deser_cnt <= 1;
			end
		endcase
		deser_last <= serdes_in[3];
	end

	reg [3:0] bitcount = 0;
	reg [9:0] bitbuffer = 0;

	// wire [2:0] deser_en_sum = deser_en[0] + deser_en[1] + deser_en[2] + deser_en[3];
	wire [2:0] deser_en_sum = deser_en[3] ? 4 : deser_en[2] ? 3 : deser_en[1] ? 2 : deser_en[0] ? 1 : 0;

	always @(posedge clk) begin
		recv_word_en <= 0;
		bitbuffer <= {deser_bits, bitbuffer} >> deser_en_sum;
		if (deser_en[0] && {deser_bits[0], bitbuffer[9:1]} == 10'b0001111100) begin
			bitcount <= deser_en_sum - 1;
			recv_word_en <= !reset;
			recv_wbits <= 10'b0001111100;

		end else
		if (&deser_en[1:0] && {deser_bits[1:0], bitbuffer[9:2]} == 10'b0001111100) begin
			bitcount <= deser_en_sum - 2;
			recv_word_en <= !reset;
			recv_wbits <= 10'b0001111100;
		end else
		if (&deser_en[2:0] && {deser_bits[2:0], bitbuffer[9:3]} == 10'b0001111100) begin
			bitcount <= deser_en_sum - 3;
			recv_word_en <= !reset;
			recv_wbits <= 10'b0001111100;
		end else
		if (&deser_en[3:0] && {deser_bits[3:0], bitbuffer[9:4]} == 10'b0001111100) begin
			bitcount <= 0;
			recv_word_en <= !reset;
			recv_wbits <= 10'b0001111100;
		end else begin
			if (bitcount + deser_en_sum >= 10) begin
				recv_word_en <= !reset;
				recv_wbits <= {deser_bits, bitbuffer} >> (10 - bitcount);
				bitcount <= bitcount + deser_en_sum - 10;
			end else
				bitcount <= bitcount + deser_en_sum;
			bitbuffer <= {deser_bits, bitbuffer} >> deser_en_sum;
		end
	end
endmodule

// Chuck Benz, Hollis, NH   Copyright (c)2002
//
// The information and description contained herein is the
// property of Chuck Benz.
//
// Permission is granted for any reuse of this information
// and description as long as this copyright notice is
// preserved.  Modifications may be made as long as this
// notice is preserved.

// per Widmer and Franaszek

// Minor modifications by Clifford Wolf for ponylink

`timescale 1 ns / 1 ps

module ponylink_decode_8b10b (datain, dispin, dataout, dispout, code_err, disp_err) ;
  input [9:0]   datain ;
  input		dispin ;
  output [8:0]	dataout ;
  output	dispout ;
  output	code_err ;
  output	disp_err ;

  wire ai = datain[0] ;
  wire bi = datain[1] ;
  wire ci = datain[2] ;
  wire di = datain[3] ;
  wire ei = datain[4] ;
  wire ii = datain[5] ;
  wire fi = datain[6] ;
  wire gi = datain[7] ;
  wire hi = datain[8] ;
  wire ji = datain[9] ;

  wire aeqb = (ai & bi) | (!ai & !bi) ;
  wire ceqd = (ci & di) | (!ci & !di) ;
  wire p22 = (ai & bi & !ci & !di) |
	     (ci & di & !ai & !bi) |
	     ( !aeqb & !ceqd) ;
  wire p13 = ( !aeqb & !ci & !di) |
	     ( !ceqd & !ai & !bi) ;
  wire p31 = ( !aeqb & ci & di) |
	     ( !ceqd & ai & bi) ;

  wire p40 = ai & bi & ci & di ;
  wire p04 = !ai & !bi & !ci & !di ;

  wire disp6a = p31 | (p22 & dispin) ; // pos disp if p22 and was pos, or p31.
   wire disp6a2 = p31 & dispin ;  // disp is ++ after 4 bits
   wire disp6a0 = p13 & ! dispin ; // -- disp after 4 bits
    
  wire disp6b = (((ei & ii & ! disp6a0) | (disp6a & (ei | ii)) | disp6a2 |
		  (ei & ii & di)) & (ei | ii | di)) ;

  // The 5B/6B decoding special cases where ABCDE != abcde

  wire p22bceeqi = p22 & bi & ci & (ei == ii) ;
  wire p22bncneeqi = p22 & !bi & !ci & (ei == ii) ;
  wire p13in = p13 & !ii ;
  wire p31i = p31 & ii ;
  wire p13dei = p13 & di & ei & ii ;
  wire p22aceeqi = p22 & ai & ci & (ei == ii) ;
  wire p22ancneeqi = p22 & !ai & !ci & (ei == ii) ;
  wire p13en = p13 & !ei ;
  wire anbnenin = !ai & !bi & !ei & !ii ;
  wire abei = ai & bi & ei & ii ;
  wire cdei = ci & di & ei & ii ;
  wire cndnenin = !ci & !di & !ei & !ii ;

  // non-zero disparity cases:
  wire p22enin = p22 & !ei & !ii ;
  wire p22ei = p22 & ei & ii ;
  //wire p13in = p12 & !ii ;
  //wire p31i = p31 & ii ;
  wire p31dnenin = p31 & !di & !ei & !ii ;
  //wire p13dei = p13 & di & ei & ii ;
  wire p31e = p31 & ei ;

  wire compa = p22bncneeqi | p31i | p13dei | p22ancneeqi | 
		p13en | abei | cndnenin ;
  wire compb = p22bceeqi | p31i | p13dei | p22aceeqi | 
		p13en | abei | cndnenin ;
  wire compc = p22bceeqi | p31i | p13dei | p22ancneeqi | 
		p13en | anbnenin | cndnenin ;
  wire compd = p22bncneeqi | p31i | p13dei | p22aceeqi |
		p13en | abei | cndnenin ;
  wire compe = p22bncneeqi | p13in | p13dei | p22ancneeqi | 
		p13en | anbnenin | cndnenin ;

  wire ao = ai ^ compa ;
  wire bo = bi ^ compb ;
  wire co = ci ^ compc ;
  wire do = di ^ compd ;
  wire eo = ei ^ compe ;

  wire feqg = (fi & gi) | (!fi & !gi) ;
  wire heqj = (hi & ji) | (!hi & !ji) ;
  wire fghj22 = (fi & gi & !hi & !ji) |
		(!fi & !gi & hi & ji) |
		( !feqg & !heqj) ;
  wire fghjp13 = ( !feqg & !hi & !ji) |
		 ( !heqj & !fi & !gi) ;
  wire fghjp31 = ( (!feqg) & hi & ji) |
		 ( !heqj & fi & gi) ;

  wire dispout = (fghjp31 | (disp6b & fghj22) | (hi & ji)) & (hi | ji) ;

  wire ko = ( (ci & di & ei & ii) | ( !ci & !di & !ei & !ii) |
		(p13 & !ei & ii & gi & hi & ji) |
		(p31 & ei & !ii & !gi & !hi & !ji)) ;

  wire alt7 =   (fi & !gi & !hi & // 1000 cases, where disp6b is 1
		 ((dispin & ci & di & !ei & !ii) | ko |
		  (dispin & !ci & di & !ei & !ii))) |
		(!fi & gi & hi & // 0111 cases, where disp6b is 0
		 (( !dispin & !ci & !di & ei & ii) | ko |
		  ( !dispin & ci & !di & ei & ii))) ;

  wire k28 = (ci & di & ei & ii) | ! (ci | di | ei | ii) ;
  // k28 with positive disp into fghi - .1, .2, .5, and .6 special cases
  wire k28p = ! (ci | di | ei | ii) ;
  wire fo = (ji & !fi & (hi | !gi | k28p)) |
	    (fi & !ji & (!hi | gi | !k28p)) |
	    (k28p & gi & hi) |
	    (!k28p & !gi & !hi) ;
  wire go = (ji & !fi & (hi | !gi | !k28p)) |
	    (fi & !ji & (!hi | gi |k28p)) |
	    (!k28p & gi & hi) |
	    (k28p & !gi & !hi) ;
  wire ho = ((ji ^ hi) & ! ((!fi & gi & !hi & ji & !k28p) | (!fi & gi & hi & !ji & k28p) | 
			    (fi & !gi & !hi & ji & !k28p) | (fi & !gi & hi & !ji & k28p))) |
	    (!fi & gi & hi & ji) | (fi & !gi & !hi & !ji) ;

  wire disp6p = (p31 & (ei | ii)) | (p22 & ei & ii) ;
  wire disp6n = (p13 & ! (ei & ii)) | (p22 & !ei & !ii) ;
  wire disp4p = fghjp31 ;
  wire disp4n = fghjp13 ;

  assign code_err = p40 | p04 | (fi & gi & hi & ji) | (!fi & !gi & !hi & !ji) |
		    (p13 & !ei & !ii) | (p31 & ei & ii) | 
		    (ei & ii & fi & gi & hi) | (!ei & !ii & !fi & !gi & !hi) | 
		    (ei & !ii & gi & hi & ji) | (!ei & ii & !gi & !hi & !ji) |
		    (!p31 & ei & !ii & !gi & !hi & !ji) |
		    (!p13 & !ei & ii & gi & hi & ji) |
		    (((ei & ii & !gi & !hi & !ji) | 
		      (!ei & !ii & gi & hi & ji)) &
		     ! ((ci & di & ei) | (!ci & !di & !ei))) |
		    (disp6p & disp4p) | (disp6n & disp4n) |
		    (ai & bi & ci & !ei & !ii & ((!fi & !gi) | fghjp13)) |
		    (!ai & !bi & !ci & ei & ii & ((fi & gi) | fghjp31)) |
		    (fi & gi & !hi & !ji & disp6p) |
		    (!fi & !gi & hi & ji & disp6n) |
		    (ci & di & ei & ii & !fi & !gi & !hi) |
		    (!ci & !di & !ei & !ii & fi & gi & hi) ;

  assign dataout = {ko, ho, go, fo, eo, do, co, bo, ao} ;

  // my disp err fires for any legal codes that violate disparity, may fire for illegal codes
   assign disp_err = ((dispin & disp6p) | (disp6n & !dispin) |
		      (dispin & !disp6n & fi & gi) |
		      (dispin & ai & bi & ci) |
		      (dispin & !disp6n & disp4p) |
		      (!dispin & !disp6p & !fi & !gi) |
		      (!dispin & !ai & !bi & !ci) |
		      (!dispin & !disp6p & disp4n) |
		      (disp6p & disp4p) | (disp6n & disp4n)) ;
endmodule

// ========================================================================================

module ponylink_encode_8b10b (datain, dispin, dataout, dispout) ;
  input [8:0]   datain ;
  input 	dispin ;  // 0 = neg disp; 1 = pos disp
  output [9:0]	dataout ;
  output	dispout ;


  wire ai = datain[0] ;
  wire bi = datain[1] ;
  wire ci = datain[2] ;
  wire di = datain[3] ;
  wire ei = datain[4] ;
  wire fi = datain[5] ;
  wire gi = datain[6] ;
  wire hi = datain[7] ;
  wire ki = datain[8] ;

  wire aeqb = (ai & bi) | (!ai & !bi) ;
  wire ceqd = (ci & di) | (!ci & !di) ;
  wire l22 = (ai & bi & !ci & !di) |
	     (ci & di & !ai & !bi) |
	     ( !aeqb & !ceqd) ;
  wire l40 = ai & bi & ci & di ;
  wire l04 = !ai & !bi & !ci & !di ;
  wire l13 = ( !aeqb & !ci & !di) |
	     ( !ceqd & !ai & !bi) ;
  wire l31 = ( !aeqb & ci & di) |
	     ( !ceqd & ai & bi) ;

  // The 5B/6B encoding

  wire ao = ai ;
  wire bo = (bi & !l40) | l04 ;
  wire co = l04 | ci | (ei & di & !ci & !bi & !ai) ;
  wire do = di & ! (ai & bi & ci) ;
  wire eo = (ei | l13) & ! (ei & di & !ci & !bi & !ai) ;
  wire io = (l22 & !ei) |
	    (ei & !di & !ci & !(ai&bi)) |  // D16, D17, D18
	    (ei & l40) |
	    (ki & ei & di & ci & !bi & !ai) | // K.28
	    (ei & !di & ci & !bi & !ai) ;

  // pds16 indicates cases where d-1 is assumed + to get our encoded value
  wire pd1s6 = (ei & di & !ci & !bi & !ai) | (!ei & !l22 & !l31) ;
  // nds16 indicates cases where d-1 is assumed - to get our encoded value
  wire nd1s6 = ki | (ei & !l22 & !l13) | (!ei & !di & ci & bi & ai) ;

  // ndos6 is pds16 cases where d-1 is + yields - disp out - all of them
  wire ndos6 = pd1s6 ;
  // pdos6 is nds16 cases where d-1 is - yields + disp out - all but one
  wire pdos6 = ki | (ei & !l22 & !l13) ;


  // some Dx.7 and all Kx.7 cases result in run length of 5 case unless
  // an alternate coding is used (referred to as Dx.A7, normal is Dx.P7)
  // specifically, D11, D13, D14, D17, D18, D19.
  wire alt7 = fi & gi & hi & (ki | 
			      (dispin ? (!ei & di & l31) : (ei & !di & l13))) ;

   
  wire fo = fi & ! alt7 ;
  wire go = gi | (!fi & !gi & !hi) ;
  wire ho = hi ;
  wire jo = (!hi & (gi ^ fi)) | alt7 ;

  // nd1s4 is cases where d-1 is assumed - to get our encoded value
  wire nd1s4 = fi & gi ;
  // pd1s4 is cases where d-1 is assumed + to get our encoded value
  wire pd1s4 = (!fi & !gi) | (ki & ((fi & !gi) | (!fi & gi))) ;

  // ndos4 is pd1s4 cases where d-1 is + yields - disp out - just some
  wire ndos4 = (!fi & !gi) ;
  // pdos4 is nd1s4 cases where d-1 is - yields + disp out 
  wire pdos4 = fi & gi & hi ;

  // only legal K codes are K28.0->.7, K23/27/29/30.7
  //	K28.0->7 is ei=di=ci=1,bi=ai=0
  //	K23 is 10111
  //	K27 is 11011
  //	K29 is 11101
  //	K30 is 11110 - so K23/27/29/30 are ei & l31
  wire illegalk = ki & 
		  (ai | bi | !ci | !di | !ei) & // not K28.0->7
		  (!fi | !gi | !hi | !ei | !l31) ; // not K23/27/29/30.7

  // now determine whether to do the complementing
  // complement if prev disp is - and pd1s6 is set, or + and nd1s6 is set
  wire compls6 = (pd1s6 & !dispin) | (nd1s6 & dispin) ;

  // disparity out of 5b6b is disp in with pdso6 and ndso6
  // pds16 indicates cases where d-1 is assumed + to get our encoded value
  // ndos6 is cases where d-1 is + yields - disp out
  // nds16 indicates cases where d-1 is assumed - to get our encoded value
  // pdos6 is cases where d-1 is - yields + disp out
  // disp toggles in all ndis16 cases, and all but that 1 nds16 case

  wire disp6 = dispin ^ (ndos6 | pdos6) ;

  wire compls4 = (pd1s4 & !disp6) | (nd1s4 & disp6) ;
  assign dispout = disp6 ^ (ndos4 | pdos4) ;

  assign dataout = {(jo ^ compls4), (ho ^ compls4),
		    (go ^ compls4), (fo ^ compls4),
		    (io ^ compls6), (eo ^ compls6),
		    (do ^ compls6), (co ^ compls6),
		    (bo ^ compls6), (ao ^ compls6)} ;
endmodule

// -----------------------------------------------------------------------
// Wrappers with extra functionality for PonyLink. Added by Clifford Wolf.

module ponylink_encode_8b10b_xtra (clk, datain, dispin, dataout, dispout);
	input clk;
	input [8:0] datain;
	input dispin;
	output [9:0] dataout;
	output dispout;
	wire [9:0] dataout_t;
	reg lastbit;

	ponylink_encode_8b10b encoder (datain, dispin, dataout_t, dispout);

	assign dataout = datain == 9'h100 ? {9'b0, ~lastbit} : dataout_t;
	always @(posedge clk) lastbit <= dataout[9];
endmodule

module ponylink_decode_8b10b_xtra (clk, reset, enable, rstdetect, datain, dataout, recv_error);
	parameter RECVRESET = 0;

	input clk, reset, enable;
	output reg rstdetect;
	input [9:0] datain;
	output [8:0] dataout;
	output recv_error;

	reg dispout_q;
	wire dispin, dispout;

	always @(posedge clk)
		if (enable) dispout_q <= dispout;
	
	assign dispin = dispout_q && (datain != 10'b0001111100);

	wire code_err, disp_err;
	assign recv_error = (code_err | disp_err) && dataout != 9'h1fc;

	ponylink_decode_8b10b decoder (datain, dispin, dataout, dispout, code_err, disp_err);

	reg [3:0] reset_state;

	always @(posedge clk) begin
		rstdetect <= 0;
		if (reset) begin
			reset_state <= 0;
		end else
		if (enable) begin 
			reset_state <= 0;
			if (!recv_error) begin
				if (dataout == 9'h1fc)
					reset_state <= 2;
				else
				if (2 <= reset_state && reset_state < 6) begin
					if (dataout == (RECVRESET ? 9'h1fd : 9'h1fe))
						reset_state <= reset_state + 1'b1;
				end else
				if (6 <= reset_state) begin
					if (dataout == reset_state)
						reset_state <= reset_state + 1'b1;
				end
			end
		end
		if (reset_state == 15) begin
			rstdetect <= 1;
			reset_state <= 0;
		end
	end
endmodule

//-----------------------------------------------------------------------------
// Copyright (C) 2009 OutputLogic.com 
// This source file may be used and distributed without restriction 
// provided that this copyright statement is not removed from the file 
// and that any derivative work contains the original copyright notice 
// and the associated disclaimer. 
// 
// THIS SOURCE FILE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS 
// OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED	
// WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE. 
//-----------------------------------------------------------------------------
// CRC module for data[7:0] ,   crc[31:0]=1+x^1+x^2+x^4+x^5+x^7+x^8+x^10+x^11+x^12+x^16+x^22+x^23+x^26+x^32;
//-----------------------------------------------------------------------------
module ponylink_crc32(
  input [7:0] data_in,
  input crc_en,
  output [31:0] crc_out,
  input rst,
  input clk);

  reg [31:0] lfsr_q,lfsr_c;

  assign crc_out = lfsr_q;

  always @(*) begin
    lfsr_c[0] = lfsr_q[24] ^ lfsr_q[30] ^ data_in[0] ^ data_in[6];
    lfsr_c[1] = lfsr_q[24] ^ lfsr_q[25] ^ lfsr_q[30] ^ lfsr_q[31] ^ data_in[0] ^ data_in[1] ^ data_in[6] ^ data_in[7];
    lfsr_c[2] = lfsr_q[24] ^ lfsr_q[25] ^ lfsr_q[26] ^ lfsr_q[30] ^ lfsr_q[31] ^ data_in[0] ^ data_in[1] ^ data_in[2] ^ data_in[6] ^ data_in[7];
    lfsr_c[3] = lfsr_q[25] ^ lfsr_q[26] ^ lfsr_q[27] ^ lfsr_q[31] ^ data_in[1] ^ data_in[2] ^ data_in[3] ^ data_in[7];
    lfsr_c[4] = lfsr_q[24] ^ lfsr_q[26] ^ lfsr_q[27] ^ lfsr_q[28] ^ lfsr_q[30] ^ data_in[0] ^ data_in[2] ^ data_in[3] ^ data_in[4] ^ data_in[6];
    lfsr_c[5] = lfsr_q[24] ^ lfsr_q[25] ^ lfsr_q[27] ^ lfsr_q[28] ^ lfsr_q[29] ^ lfsr_q[30] ^ lfsr_q[31] ^ data_in[0] ^ data_in[1] ^ data_in[3] ^ data_in[4] ^ data_in[5] ^ data_in[6] ^ data_in[7];
    lfsr_c[6] = lfsr_q[25] ^ lfsr_q[26] ^ lfsr_q[28] ^ lfsr_q[29] ^ lfsr_q[30] ^ lfsr_q[31] ^ data_in[1] ^ data_in[2] ^ data_in[4] ^ data_in[5] ^ data_in[6] ^ data_in[7];
    lfsr_c[7] = lfsr_q[24] ^ lfsr_q[26] ^ lfsr_q[27] ^ lfsr_q[29] ^ lfsr_q[31] ^ data_in[0] ^ data_in[2] ^ data_in[3] ^ data_in[5] ^ data_in[7];
    lfsr_c[8] = lfsr_q[0] ^ lfsr_q[24] ^ lfsr_q[25] ^ lfsr_q[27] ^ lfsr_q[28] ^ data_in[0] ^ data_in[1] ^ data_in[3] ^ data_in[4];
    lfsr_c[9] = lfsr_q[1] ^ lfsr_q[25] ^ lfsr_q[26] ^ lfsr_q[28] ^ lfsr_q[29] ^ data_in[1] ^ data_in[2] ^ data_in[4] ^ data_in[5];
    lfsr_c[10] = lfsr_q[2] ^ lfsr_q[24] ^ lfsr_q[26] ^ lfsr_q[27] ^ lfsr_q[29] ^ data_in[0] ^ data_in[2] ^ data_in[3] ^ data_in[5];
    lfsr_c[11] = lfsr_q[3] ^ lfsr_q[24] ^ lfsr_q[25] ^ lfsr_q[27] ^ lfsr_q[28] ^ data_in[0] ^ data_in[1] ^ data_in[3] ^ data_in[4];
    lfsr_c[12] = lfsr_q[4] ^ lfsr_q[24] ^ lfsr_q[25] ^ lfsr_q[26] ^ lfsr_q[28] ^ lfsr_q[29] ^ lfsr_q[30] ^ data_in[0] ^ data_in[1] ^ data_in[2] ^ data_in[4] ^ data_in[5] ^ data_in[6];
    lfsr_c[13] = lfsr_q[5] ^ lfsr_q[25] ^ lfsr_q[26] ^ lfsr_q[27] ^ lfsr_q[29] ^ lfsr_q[30] ^ lfsr_q[31] ^ data_in[1] ^ data_in[2] ^ data_in[3] ^ data_in[5] ^ data_in[6] ^ data_in[7];
    lfsr_c[14] = lfsr_q[6] ^ lfsr_q[26] ^ lfsr_q[27] ^ lfsr_q[28] ^ lfsr_q[30] ^ lfsr_q[31] ^ data_in[2] ^ data_in[3] ^ data_in[4] ^ data_in[6] ^ data_in[7];
    lfsr_c[15] = lfsr_q[7] ^ lfsr_q[27] ^ lfsr_q[28] ^ lfsr_q[29] ^ lfsr_q[31] ^ data_in[3] ^ data_in[4] ^ data_in[5] ^ data_in[7];
    lfsr_c[16] = lfsr_q[8] ^ lfsr_q[24] ^ lfsr_q[28] ^ lfsr_q[29] ^ data_in[0] ^ data_in[4] ^ data_in[5];
    lfsr_c[17] = lfsr_q[9] ^ lfsr_q[25] ^ lfsr_q[29] ^ lfsr_q[30] ^ data_in[1] ^ data_in[5] ^ data_in[6];
    lfsr_c[18] = lfsr_q[10] ^ lfsr_q[26] ^ lfsr_q[30] ^ lfsr_q[31] ^ data_in[2] ^ data_in[6] ^ data_in[7];
    lfsr_c[19] = lfsr_q[11] ^ lfsr_q[27] ^ lfsr_q[31] ^ data_in[3] ^ data_in[7];
    lfsr_c[20] = lfsr_q[12] ^ lfsr_q[28] ^ data_in[4];
    lfsr_c[21] = lfsr_q[13] ^ lfsr_q[29] ^ data_in[5];
    lfsr_c[22] = lfsr_q[14] ^ lfsr_q[24] ^ data_in[0];
    lfsr_c[23] = lfsr_q[15] ^ lfsr_q[24] ^ lfsr_q[25] ^ lfsr_q[30] ^ data_in[0] ^ data_in[1] ^ data_in[6];
    lfsr_c[24] = lfsr_q[16] ^ lfsr_q[25] ^ lfsr_q[26] ^ lfsr_q[31] ^ data_in[1] ^ data_in[2] ^ data_in[7];
    lfsr_c[25] = lfsr_q[17] ^ lfsr_q[26] ^ lfsr_q[27] ^ data_in[2] ^ data_in[3];
    lfsr_c[26] = lfsr_q[18] ^ lfsr_q[24] ^ lfsr_q[27] ^ lfsr_q[28] ^ lfsr_q[30] ^ data_in[0] ^ data_in[3] ^ data_in[4] ^ data_in[6];
    lfsr_c[27] = lfsr_q[19] ^ lfsr_q[25] ^ lfsr_q[28] ^ lfsr_q[29] ^ lfsr_q[31] ^ data_in[1] ^ data_in[4] ^ data_in[5] ^ data_in[7];
    lfsr_c[28] = lfsr_q[20] ^ lfsr_q[26] ^ lfsr_q[29] ^ lfsr_q[30] ^ data_in[2] ^ data_in[5] ^ data_in[6];
    lfsr_c[29] = lfsr_q[21] ^ lfsr_q[27] ^ lfsr_q[30] ^ lfsr_q[31] ^ data_in[3] ^ data_in[6] ^ data_in[7];
    lfsr_c[30] = lfsr_q[22] ^ lfsr_q[28] ^ lfsr_q[31] ^ data_in[4] ^ data_in[7];
    lfsr_c[31] = lfsr_q[23] ^ lfsr_q[29] ^ data_in[5];

  end // always

  always @(posedge clk) begin
    if(rst) begin
      lfsr_q <= {32{1'b1}};
    end
    else begin
      lfsr_q <= crc_en ? lfsr_c : lfsr_q;
    end
  end // always
endmodule // crc
module main (
	input        clk,
	input        serdes_in,
	output       serdes_out,
	output       serdes_en
);
	wire clk;
	wire resetn_out;
	wire linkerror;
	wire linkready;
	wire mode_recv;
	wire mode_send;
	wire [7:0] gpio_i = 0;
	wire [7:0] gpio_o;
	wire [4:0] in_tdata = 0;
	wire [3:0] in_tuser = 0;
	wire in_tvalid = 1;
	wire in_tlast = 0;
	wire in_tready;
	wire [4:0] out_tdata;
	wire [3:0] out_tuser;
	wire out_tvalid;
	wire out_tlast;
	wire out_tready = 1;
	wire serdes_in;
	wire serdes_out;
	wire serdes_en;

	ponylink_slave #(
		.SERDES_REG_IN(1),
		.SERDES_REG_OUT(1),
		.M2S_TDATA_WIDTH(8),
		.M2S_TUSER_WIDTH(4),
		.S2M_TDATA_WIDTH(8),
		.S2M_TUSER_WIDTH(4),
		.MASTER_RECV_DELAY(8),
		.SLAVE_RECV_DELAY(8),
		.MASTER_SEND_DELAY(32),
		.SLAVE_SEND_DELAY(32),
		.MASTER_PARBITS(1),
		.SLAVE_PARBITS(1),
		.MASTER_PKTLEN(8),
		.SLAVE_PKTLEN(8),
		.MASTER_TIMINGS(80'h2e241a100618130e0904),
		.SLAVE_TIMINGS (80'h05040302010907050301)
	) uut (
		.clk       (clk       ),
		.resetn_out(resetn_out),
		.linkerror (linkerror ),
		.linkready (linkready ),
		.mode_recv (mode_recv ),
		.mode_send (mode_send ),
		.gpio_i    (gpio_i    ),
		.gpio_o    (gpio_o    ),
		.in_tdata  (in_tdata  ),
		.in_tuser  (in_tuser  ),
		.in_tvalid (in_tvalid ),
		.in_tlast  (in_tlast  ),
		.in_tready (in_tready ),
		.out_tdata (out_tdata ),
		.out_tuser (out_tuser ),
		.out_tvalid(out_tvalid),
		.out_tlast (out_tlast ),
		.out_tready(out_tready),
		.serdes_in (serdes_in ),
		.serdes_out(serdes_out),
		.serdes_en (serdes_en )
	);

	localparam xfer_maxlen = 229;
	reg [xfer_maxlen-1:0] xfer_shreg = 0;
	always @(posedge clk) begin
		xfer_shreg <= {xfer_shreg, serdes_en};
		assert (!(&xfer_shreg));
	end
endmodule
