#!/bin/bash

readonly scriptdir=`dirname $(readlink -f $0)`

readonly toolsdir="$scriptdir/../tools/"

readonly boolector="$toolsdir/boolector"
readonly btormc="$toolsdir/btormc"
readonly yices="$toolsdir/yices"

readonly benchmarks_btor_sel="$scriptdir/benchmarks_btor_sel"
readonly benchmarks_smt2_sel="$scriptdir/benchmarks_smt2_sel"
readonly benchmarks_btor_all="$scriptdir/benchmarks_btor_all"
readonly benchmarks_smt2_all="$scriptdir/benchmarks_smt2_all"

benchmarks_btor="$benchmarks_btor_sel"
benchmarks_smt2="$benchmarks_smt2_sel"

tmpdir="$scriptdir/paper-results"

workdir_btormc=unknown
workdir_boolector=unknown
workdir_yices=unknown

readonly runsdir="$scriptdir/runs"

readonly timelimit=3600
readonly spacelimit=7000

readonly runlim="$toolsdir/runlim"
readonly runlim_opts="--time-limit=$timelimit --space-limit=$spacelimit"

readonly compare_script="$toolsdir/compare-smt-runs.py"

trap "rm -rf $tmpdir" SIGHUP SIGINT SIGTERM 

function die
{
  echo "** $(basename $0): $1"
  exit 1
}

function msg
{
  echo "[CAV18] $1"
}

runall=false

while [ $# -gt 0 ]
do
  case $1 in
    -h|--help)
      echo -n "usage: $(basename $0)"
      echo -e " [ -h | --help | -a ] [<benchmarks-file>]"
      echo -n " -a        Run all benchmarks. Default is to run a "
      echo -e "selection of 3 benchmarks."
      exit 0 ;;
    -a) runall=true ;;
    -*|--*)
      die "invalid option, try '-h'" ;;
    *) break ;;
  esac
  shift
done

if [ $# -gt 0 ]; then
  die "invalid number of arguments, try '-h'"
fi

# check if binaries are available
if [ ! -f "$boolector" ]
then
  die "could not find boolector binary (not built?)"
fi
if [ ! -f "$btormc" ]
then 
  die "could not find btormc binary (not built?)"
fi
if [ ! -f "$yices" ]
then 
  die "could not find yices binary"
fi

# check if benchmarks files are available
if [ ! -f "$benchmarks_btor" ]
then
  die "could not find btor benchmarks file"
fi
if [ ! -f "$benchmarks_smt2" ]
then
  die "could not find smt2 benchmarks file"
fi


# create working directories (for log files and other output)
# and run experiments


msg "Running experiments ..."
msg 
msg "\"New Word-Level Model Checking Format, Word-Level Model Checker and Bit-Vector Solver\""
msg "Aina Niemetz, Mathias Preiner, Clifford Wolf and Armin Biere"
msg 

# 1) btormc
benchmarks_file="$benchmarks_btor"
msg "1 .. Running btormc on `basename $benchmarks_file`"
workdir_btormc=$scriptdir/`date +%Y-%m-%d-``basename $benchmarks_file`-btormc
workdir=$workdir_btormc
mkdir -p $workdir
solver="$btormc"
for f in `cat $benchmarks_file`
do
  file=`echo $f | cut -d ';' -f 1`
  opts="-kmax `echo $f | cut -d ';' -f 2`"
  g=`basename $file`
  out=`echo ${g%.*}`
  log="$workdir/${out}.log"
  err="$workdir/${out}.err"
  echo -n "c launch: " > $log
  echo -n "$benchmarks_file $workdir $runlim_opts " >> $log
  echo -ne "$solver $opts\n" >> $log
  echo "c benchmark: $f" >> $log
  echo "c solver: $solver" >> $log
  echo "c solver options: $opts" >> $log
  $runlim $runlim_opts $solver $opts $file >> $log 2> $err
done

# 2) boolector
benchmarks_file="$benchmarks_smt2"
msg "2 .. Running boolector on `basename $benchmarks_file`"
workdir_boolector=$scriptdir/`date +%Y-%m-%d-``basename $benchmarks_file`-boolector
workdir=$workdir_boolector
mkdir -p $workdir
solver="$boolector"
opts="-i -v"
for f in `cat $benchmarks_file`
do
  g=`basename $f`
  out=`echo ${g%-unrolled.*}`
  log="$workdir/${out}.log"
  err="$workdir/${out}.err"
  echo -n "c launch: " > $log
  echo -n "$benchmarks_file $workdir $runlim_opts " >> $log
  echo -ne "$solver $opts\n" >> $log
  echo "c benchmark: $f" >> $log
  echo "c solver: $solver" >> $log
  echo "c solver options: $opts" >> $log
  $runlim $runlim_opts $solver $opts $f >> $log 2> $err
done

# 3) yices
benchmarks_file="$benchmarks_smt2"
msg "3 .. Running yices on `basename $benchmarks_file`"
workdir_yices=$scriptdir/`date +%Y-%m-%d-``basename $benchmarks_file`-yices
workdir=$workdir_yices
mkdir -p $workdir
solver="$yices"
opts="--incremental"
for f in `cat $benchmarks_file`
do
  g=`basename $f`
  out=`echo ${g%-unrolled.*}`
  log="$workdir/${out}.log"
  err="$workdir/${out}.err"
  echo -n "c launch: " > $log
  echo -n "$benchmarks_file $workdir $runlim_opts " >> $log
  echo -ne "$solver $opts\n" >> $log
  echo "c benchmark: $f" >> $log
  echo "c solver: $solver" >> $log
  echo "c solver options: $opts" >> $log
  $runlim $runlim_opts $solver $opts $f >> $log 2> $err
done



## show statistics
#msg 
#msg "Show statistics ..."

# 1) btormc
workdir=$workdir_btormc
a=unknown
while [[ $a != "y" && $a != "n" ]]
do
  msg
  msg "1 .. Show statistics for btormc: y/n"
  read a
  if [[ $a == "n" ]]; then break; fi
  if [[ $a == "y" ]]
  then
    msg
    msg "Compare current run '`basename $workdir`' with provided log files from paper ('`basename $tmpdir`')..."
    msg
    if [ -d "$tmpdir" -o -f "$tmpdir" ]; then
      die "tmp dir '$tmpdir' does already exist"
    fi
    mkdir -p $tmpdir
    for f in $workdir/*
    do
      cp "$runsdir/btormc/`basename $f`" $tmpdir
    done
    $compare_script --no-colors -s time_cpu $workdir $tmpdir
    rm -rf $tmpdir
    break
  fi
done

# 2) boolector
workdir=$workdir_boolector
a=unknown
while [[ $a != "y" && $a != "n" ]]
do
  msg
  msg "2 .. Show statistics for boolector: y/n"
  read a
  if [[ $a == "n" ]]; then break; fi
  if [[ $a == "y" ]]
  then
    msg
    msg "Compare current run '`basename $workdir`' with provided log files from paper ('`basename $tmpdir`')..."
    msg
    if [ -d "$tmpdir" -o -f "$tmpdir" ]; then
      die "tmp dir '$tmpdir' does already exist"
    fi
    mkdir -p $tmpdir
    for f in $workdir/*
    do
      cp "$runsdir/boolector/`basename $f`" $tmpdir
    done
    $compare_script --no-colors -s time_cpu $workdir $tmpdir
    rm -rf $tmpdir
    break
  fi
done

# 3) yices
workdir=$workdir_yices
a=unknown
while [[ $a != "y" && $a != "n" ]]
do
  msg
  msg "3 .. Show statistics for yices: y/n"
  read a
  if [[ $a == "n" ]]; then break; fi
  if [[ $a == "y" ]]
  then
    msg
    msg "Compare current run '`basename $workdir`' with provided log files from paper ('`basename $tmpdir`')..."
    msg
    if [ -d "$tmpdir" -o -f "$tmpdir" ]; then
      die "tmp dir '$tmpdir' does already exist"
    fi
    mkdir -p $tmpdir
    for f in $workdir/*
    do
      cp "$runsdir/yices/`basename $f`" $tmpdir
    done
    $compare_script --no-colors -s time_cpu $workdir $tmpdir
    rm -rf $tmpdir
    break
  fi
done

